# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=too-many-lines
import abc
import atexit
import concurrent.futures
import json
import logging
import os
import threading
import traceback
import typing
import weakref
from collections.abc import (
    Callable,
    Iterator,
    Mapping,
    MutableMapping,
    Sequence,
)
from dataclasses import dataclass
from os import environ
from time import time_ns
from types import MappingProxyType, TracebackType
from typing import (
    Any,
)
from warnings import filterwarnings

from typing_extensions import deprecated

from opentelemetry import context as context_api
from opentelemetry import metrics as metrics_api
from opentelemetry import trace as trace_api
from opentelemetry.attributes import BoundedAttributes
from opentelemetry.sdk import util
from opentelemetry.sdk.environment_variables import (
    OTEL_ATTRIBUTE_COUNT_LIMIT,
    OTEL_ATTRIBUTE_VALUE_LENGTH_LIMIT,
    OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT,
    OTEL_LINK_ATTRIBUTE_COUNT_LIMIT,
    OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED,
    OTEL_SDK_DISABLED,
    OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT,
    OTEL_SPAN_ATTRIBUTE_VALUE_LENGTH_LIMIT,
    OTEL_SPAN_EVENT_COUNT_LIMIT,
    OTEL_SPAN_LINK_COUNT_LIMIT,
)
from opentelemetry.sdk.environment_variables._internal import (
    parse_boolean_environment_variable,
)
from opentelemetry.sdk.resources import (
    Resource,
    _get_process_dependent_resource,
)
from opentelemetry.sdk.trace import sampling
from opentelemetry.sdk.trace._tracer_metrics import create_tracer_metrics
from opentelemetry.sdk.trace.id_generator import IdGenerator, RandomIdGenerator
from opentelemetry.sdk.util import BoundedList
from opentelemetry.sdk.util._configurator import RuleBasedConfigurator
from opentelemetry.sdk.util.instrumentation import (
    InstrumentationInfo,
    InstrumentationScope,
)
from opentelemetry.semconv.attributes.exception_attributes import (
    EXCEPTION_ESCAPED,
    EXCEPTION_MESSAGE,
    EXCEPTION_STACKTRACE,
    EXCEPTION_TYPE,
)
from opentelemetry.trace import NoOpTracer, SpanContext
from opentelemetry.trace.status import Status, StatusCode
from opentelemetry.util import types
from opentelemetry.util._decorator import _agnosticcontextmanager

logger = logging.getLogger(__name__)

_DEFAULT_OTEL_ATTRIBUTE_COUNT_LIMIT = 128
_DEFAULT_OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT = 128
_DEFAULT_OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT = 128
_DEFAULT_OTEL_LINK_ATTRIBUTE_COUNT_LIMIT = 128
_DEFAULT_OTEL_SPAN_EVENT_COUNT_LIMIT = 128
_DEFAULT_OTEL_SPAN_LINK_COUNT_LIMIT = 128


_ENV_VALUE_UNSET = ""


class SpanProcessor:
    """Interface which allows hooks for SDK's `Span` start and end method
    invocations.

    Span processors can be registered directly using
    :func:`TracerProvider.add_span_processor` and they are invoked
    in the same order as they were registered.
    """

    def on_start(
        self,
        span: "Span",
        parent_context: context_api.Context | None = None,
    ) -> None:
        """Called when a :class:`opentelemetry.trace.Span` is started.

        This method is called synchronously on the thread that starts the
        span, therefore it should not block or throw an exception.

        Args:
            span: The :class:`opentelemetry.trace.Span` that just started.
            parent_context: The parent context of the span that just started.
        """

    def _on_ending(self, span: "Span") -> None:
        """Called when a :class:`opentelemetry.trace.Span` is ending.

        This method is called synchronously on the thread that ends the
        span, therefore it should not block or throw an exception.

        Args:
            span: The :class:`opentelemetry.trace.Span` that is ending.
        """

    def on_end(self, span: "ReadableSpan") -> None:
        """Called when a :class:`opentelemetry.trace.Span` is ended.

        This method is called synchronously on the thread that ends the
        span, therefore it should not block or throw an exception.

        Args:
            span: The :class:`opentelemetry.trace.Span` that just ended.
        """

    def shutdown(self) -> None:
        """Called when a :class:`opentelemetry.sdk.trace.TracerProvider` is shutdown."""

    def force_flush(self, timeout_millis: int = 30000) -> bool:  # pylint: disable=no-self-use
        """Export all ended spans to the configured Exporter that have not yet
        been exported.

        Args:
            timeout_millis: The maximum amount of time to wait for spans to be
                exported.

        Returns:
            False if the timeout is exceeded, True otherwise.
        """
        return True


# Temporary fix until https://github.com/PyCQA/pylint/issues/4098 is resolved
# pylint:disable=no-member
class SynchronousMultiSpanProcessor(SpanProcessor):
    """Implementation of class:`SpanProcessor` that forwards all received
    events to a list of span processors sequentially.

    The underlying span processors are called in sequential order as they were
    added.
    """

    _span_processors: tuple[SpanProcessor, ...]

    def __init__(self):
        # use a tuple to avoid race conditions when adding a new span and
        # iterating through it on "on_start" and "on_end".
        self._span_processors = ()
        self._lock = threading.Lock()

    def add_span_processor(self, span_processor: SpanProcessor) -> None:
        """Adds a SpanProcessor to the list handled by this instance."""
        with self._lock:
            self._span_processors += (span_processor,)

    def on_start(
        self,
        span: "Span",
        parent_context: context_api.Context | None = None,
    ) -> None:
        for sp in self._span_processors:
            sp.on_start(span, parent_context=parent_context)

    def _on_ending(self, span: "Span") -> None:
        for sp in self._span_processors:
            # pylint: disable=protected-access
            sp._on_ending(span)

    def on_end(self, span: "ReadableSpan") -> None:
        for sp in self._span_processors:
            sp.on_end(span)

    def shutdown(self) -> None:
        """Sequentially shuts down all underlying span processors."""
        for sp in self._span_processors:
            sp.shutdown()

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        """Sequentially calls force_flush on all underlying
        :class:`SpanProcessor`

        Args:
            timeout_millis: The maximum amount of time over all span processors
                to wait for spans to be exported. In case the first n span
                processors exceeded the timeout followup span processors will be
                skipped.

        Returns:
            True if all span processors flushed their spans within the
            given timeout, False otherwise.
        """
        deadline_ns = time_ns() + timeout_millis * 1000000
        all_flushed = True
        for sp in self._span_processors:
            current_time_ns = time_ns()
            if current_time_ns >= deadline_ns:
                return False

            if (
                sp.force_flush((deadline_ns - current_time_ns) // 1000000)
                is False
            ):
                all_flushed = False

        return all_flushed


class ConcurrentMultiSpanProcessor(SpanProcessor):
    """Implementation of :class:`SpanProcessor` that forwards all received
    events to a list of span processors in parallel.

    Calls to the underlying span processors are forwarded in parallel by
    submitting them to a thread pool executor and waiting until each span
    processor finished its work.

    Args:
        num_threads: The number of threads managed by the thread pool executor
            and thus defining how many span processors can work in parallel.
    """

    _span_processors: tuple[SpanProcessor, ...]

    def __init__(self, num_threads: int = 2):
        # use a tuple to avoid race conditions when adding a new span and
        # iterating through it on "on_start" and "on_end".
        self._span_processors = ()
        self._lock = threading.Lock()
        self._init_executor(num_threads)
        if hasattr(os, "register_at_fork"):
            # Only the main thread is kept in forked processed, the executor
            # needs to be re-instantiated to get a fresh pool of threads:
            weak_reinit = weakref.WeakMethod(self._init_executor)

            def _after_in_child() -> None:
                reinit = weak_reinit()
                if reinit is not None:
                    reinit(num_threads)

            os.register_at_fork(after_in_child=_after_in_child)

    def _init_executor(self, num_threads: int) -> None:
        self._executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=num_threads
        )

    def add_span_processor(self, span_processor: SpanProcessor) -> None:
        """Adds a SpanProcessor to the list handled by this instance."""
        with self._lock:
            self._span_processors += (span_processor,)

    def _submit_and_await(
        self,
        func: Callable[[SpanProcessor], Callable[..., None]],
        *args: Any,
        **kwargs: Any,
    ):
        futures = []
        for sp in self._span_processors:
            future = self._executor.submit(func(sp), *args, **kwargs)
            futures.append(future)
        for future in futures:
            future.result()

    def on_start(
        self,
        span: "Span",
        parent_context: context_api.Context | None = None,
    ) -> None:
        self._submit_and_await(
            lambda sp: sp.on_start, span, parent_context=parent_context
        )

    def _on_ending(self, span: "Span") -> None:
        # pylint: disable=protected-access
        self._submit_and_await(lambda sp: sp._on_ending, span)

    def on_end(self, span: "ReadableSpan") -> None:
        self._submit_and_await(lambda sp: sp.on_end, span)

    def shutdown(self) -> None:
        """Shuts down all underlying span processors in parallel."""
        self._submit_and_await(lambda sp: sp.shutdown)

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        """Calls force_flush on all underlying span processors in parallel.

        Args:
            timeout_millis: The maximum amount of time to wait for spans to be
                exported.

        Returns:
            True if all span processors flushed their spans within the given
            timeout, False otherwise.
        """
        futures = []
        for sp in self._span_processors:
            future = self._executor.submit(sp.force_flush, timeout_millis)
            futures.append(future)

        timeout_sec = timeout_millis / 1e3
        done_futures, not_done_futures = concurrent.futures.wait(
            futures, timeout_sec
        )
        if not_done_futures:
            return False

        for future in done_futures:
            if future.result() is False:
                return False

        return True


class EventBase(abc.ABC):
    def __init__(self, name: str, timestamp: int | None = None) -> None:
        self._name = name
        if timestamp is None:
            self._timestamp = time_ns()
        else:
            self._timestamp = timestamp

    @property
    def name(self) -> str:
        return self._name

    @property
    def timestamp(self) -> int:
        return self._timestamp

    @property
    @abc.abstractmethod
    def attributes(self) -> types.Attributes:
        pass


class Event(EventBase):
    """A text annotation with a set of attributes. The attributes of an event
    are immutable.

    Args:
        name: Name of the event.
        attributes: Attributes of the event.
        timestamp: Timestamp of the event. If `None` it will filled
            automatically.
    """

    def __init__(
        self,
        name: str,
        attributes: types.Attributes = None,
        timestamp: int | None = None,
        limit: int | None = _DEFAULT_OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT,
    ) -> None:
        super().__init__(name, timestamp)
        self._attributes = attributes

    @property
    def attributes(self) -> types.Attributes:
        return self._attributes

    @property
    def dropped_attributes(self) -> int:
        if isinstance(self._attributes, BoundedAttributes):
            return self._attributes.dropped
        return 0


def _check_span_ended(func):
    def wrapper(self, *args, **kwargs):
        already_ended = False
        with self._lock:  # pylint: disable=protected-access
            if self._end_time is None:  # pylint: disable=protected-access
                func(self, *args, **kwargs)
            else:
                already_ended = True

        if already_ended:
            logger.warning("Tried calling %s on an ended span.", func.__name__)

    return wrapper


def _is_valid_link(context: SpanContext, attributes: types.Attributes) -> bool:
    return bool(
        context and (context.is_valid or (attributes or context.trace_state))
    )


class ReadableSpan:
    """Provides read-only access to span attributes.

    Users should NOT be creating these objects directly. `ReadableSpan`s are created as
    a direct result from using the tracing pipeline via the `Tracer`.

    """

    def __init__(
        self,
        name: str,
        context: trace_api.SpanContext | None = None,
        parent: trace_api.SpanContext | None = None,
        resource: Resource | None = None,
        attributes: types.Attributes = None,
        events: Sequence[Event] = (),
        links: Sequence[trace_api.Link] = (),
        kind: trace_api.SpanKind = trace_api.SpanKind.INTERNAL,
        instrumentation_info: InstrumentationInfo | None = None,
        status: Status = Status(StatusCode.UNSET),
        start_time: int | None = None,
        end_time: int | None = None,
        instrumentation_scope: InstrumentationScope | None = None,
    ) -> None:
        self._name = name
        self._context = context
        self._kind = kind
        self._instrumentation_info = instrumentation_info
        self._instrumentation_scope = instrumentation_scope
        self._parent = parent
        self._start_time = start_time
        self._end_time = end_time
        self._attributes = attributes
        self._events = events
        self._links = links
        if resource is None:
            self._resource = Resource.create({})
        else:
            self._resource = resource
        self._status = status

    @property
    def dropped_attributes(self) -> int:
        if isinstance(self._attributes, BoundedAttributes):
            return self._attributes.dropped
        return 0

    @property
    def dropped_events(self) -> int:
        if isinstance(self._events, BoundedList):
            return self._events.dropped
        return 0

    @property
    def dropped_links(self) -> int:
        if isinstance(self._links, BoundedList):
            return self._links.dropped
        return 0

    @property
    def name(self) -> str:
        return self._name

    def get_span_context(self) -> trace_api.SpanContext | None:
        return self._context

    @property
    def context(self):
        return self._context

    @property
    def kind(self) -> trace_api.SpanKind:
        return self._kind

    @property
    def parent(self) -> trace_api.SpanContext | None:
        return self._parent

    @property
    def start_time(self) -> int | None:
        return self._start_time

    @property
    def end_time(self) -> int | None:
        return self._end_time

    @property
    def status(self) -> trace_api.Status:
        return self._status

    @property
    def attributes(self) -> types.Attributes:
        return MappingProxyType(self._attributes or {})

    @property
    def events(self) -> Sequence[Event]:
        return tuple(self._events)

    @property
    def links(self) -> Sequence[trace_api.Link]:
        return tuple(self._links)

    @property
    def resource(self) -> Resource:
        return self._resource

    @property
    @deprecated(
        "You should use instrumentation_scope. Deprecated since version 1.11.1."
    )
    def instrumentation_info(self) -> InstrumentationInfo | None:
        return self._instrumentation_info

    @property
    def instrumentation_scope(self) -> InstrumentationScope | None:
        return self._instrumentation_scope

    def to_json(self, indent: int | None = 4):
        parent_id = None
        if self.parent is not None:
            parent_id = f"0x{trace_api.format_span_id(self.parent.span_id)}"

        start_time = None
        if self._start_time:
            start_time = util.ns_to_iso_str(self._start_time)

        end_time = None
        if self._end_time:
            end_time = util.ns_to_iso_str(self._end_time)

        status = {
            "status_code": str(self._status.status_code.name),
        }
        if self._status.description:
            status["description"] = self._status.description

        f_span = {
            "name": self._name,
            "context": (
                self._format_context(self._context) if self._context else None
            ),
            "kind": str(self.kind),
            "parent_id": parent_id,
            "start_time": start_time,
            "end_time": end_time,
            "status": status,
            "attributes": self._format_attributes(self._attributes),
            "events": self._format_events(self._events),
            "links": self._format_links(self._links),
            "resource": json.loads(self.resource.to_json()),
        }

        return json.dumps(f_span, indent=indent)

    @staticmethod
    def _format_context(context: SpanContext) -> dict[str, str]:
        return {
            "trace_id": f"0x{trace_api.format_trace_id(context.trace_id)}",
            "span_id": f"0x{trace_api.format_span_id(context.span_id)}",
            "trace_state": repr(context.trace_state),
        }

    @staticmethod
    def _format_attributes(
        attributes: types.Attributes,
    ) -> dict[str, Any] | None:
        if attributes is not None and not isinstance(attributes, dict):
            return dict(attributes)
        return attributes

    @staticmethod
    def _format_events(events: Sequence[Event]) -> list[dict[str, Any]]:
        return [
            {
                "name": event.name,
                "timestamp": util.ns_to_iso_str(event.timestamp),
                "attributes": Span._format_attributes(  # pylint: disable=protected-access
                    event.attributes
                ),
            }
            for event in events
        ]

    @staticmethod
    def _format_links(links: Sequence[trace_api.Link]) -> list[dict[str, Any]]:
        return [
            {
                "context": Span._format_context(  # pylint: disable=protected-access
                    link.context
                ),
                "attributes": Span._format_attributes(  # pylint: disable=protected-access
                    link.attributes
                ),
            }
            for link in links
        ]


class SpanLimits:
    """The limits that should be enforce on recorded data such as events, links, attributes etc.

    This class does not enforce any limits itself. It only provides an a way read limits from env,
    default values and from user provided arguments.

    All limit arguments must be either a non-negative integer, ``None`` or ``SpanLimits.UNSET``.

    - All limit arguments are optional.
    - If a limit argument is not set, the class will try to read its value from the corresponding
      environment variable.
    - If the environment variable is not set, the default value, if any, will be used.

    Limit precedence:

    - If a model specific limit is set, it will be used.
    - Else if the corresponding global limit is set, it will be used.
    - Else if the model specific limit has a default value, the default value will be used.
    - Else if the global limit has a default value, the default value will be used.

    Args:
        max_attributes: Maximum number of attributes that can be added to a span, event, and link.
            Environment variable: OTEL_ATTRIBUTE_COUNT_LIMIT
            Default: {_DEFAULT_ATTRIBUTE_COUNT_LIMIT}
        max_events: Maximum number of events that can be added to a Span.
            Environment variable: OTEL_SPAN_EVENT_COUNT_LIMIT
            Default: {_DEFAULT_SPAN_EVENT_COUNT_LIMIT}
        max_links: Maximum number of links that can be added to a Span.
            Environment variable: OTEL_SPAN_LINK_COUNT_LIMIT
            Default: {_DEFAULT_SPAN_LINK_COUNT_LIMIT}
        max_span_attributes: Maximum number of attributes that can be added to a Span.
            Environment variable: OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT
            Default: {_DEFAULT_OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT}
        max_event_attributes: Maximum number of attributes that can be added to an Event.
            Default: {_DEFAULT_OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT}
        max_link_attributes: Maximum number of attributes that can be added to a Link.
            Default: {_DEFAULT_OTEL_LINK_ATTRIBUTE_COUNT_LIMIT}
        max_attribute_length: Maximum length an attribute value can have. Values longer than
            the specified length will be truncated.
        max_span_attribute_length: Maximum length a span attribute value can have. Values longer than
            the specified length will be truncated.
    """

    UNSET = -1

    def __init__(
        self,
        max_attributes: int | None = None,
        max_events: int | None = None,
        max_links: int | None = None,
        max_span_attributes: int | None = None,
        max_event_attributes: int | None = None,
        max_link_attributes: int | None = None,
        max_attribute_length: int | None = None,
        max_span_attribute_length: int | None = None,
    ):
        # span events and links count
        self.max_events = self._from_env_if_absent(
            max_events,
            OTEL_SPAN_EVENT_COUNT_LIMIT,
            _DEFAULT_OTEL_SPAN_EVENT_COUNT_LIMIT,
        )
        self.max_links = self._from_env_if_absent(
            max_links,
            OTEL_SPAN_LINK_COUNT_LIMIT,
            _DEFAULT_OTEL_SPAN_LINK_COUNT_LIMIT,
        )

        # attribute count
        global_max_attributes = self._from_env_if_absent(
            max_attributes, OTEL_ATTRIBUTE_COUNT_LIMIT
        )
        self.max_attributes = (
            global_max_attributes
            if global_max_attributes is not None
            else _DEFAULT_OTEL_ATTRIBUTE_COUNT_LIMIT
        )

        self.max_span_attributes = self._from_env_if_absent(
            max_span_attributes,
            OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT,
            (
                global_max_attributes
                if global_max_attributes is not None
                else _DEFAULT_OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT
            ),
        )
        self.max_event_attributes = self._from_env_if_absent(
            max_event_attributes,
            OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT,
            (
                global_max_attributes
                if global_max_attributes is not None
                else _DEFAULT_OTEL_EVENT_ATTRIBUTE_COUNT_LIMIT
            ),
        )
        self.max_link_attributes = self._from_env_if_absent(
            max_link_attributes,
            OTEL_LINK_ATTRIBUTE_COUNT_LIMIT,
            (
                global_max_attributes
                if global_max_attributes is not None
                else _DEFAULT_OTEL_LINK_ATTRIBUTE_COUNT_LIMIT
            ),
        )

        # attribute length
        self.max_attribute_length = self._from_env_if_absent(
            max_attribute_length,
            OTEL_ATTRIBUTE_VALUE_LENGTH_LIMIT,
        )
        self.max_span_attribute_length = self._from_env_if_absent(
            max_span_attribute_length,
            OTEL_SPAN_ATTRIBUTE_VALUE_LENGTH_LIMIT,
            # use global attribute length limit as default
            self.max_attribute_length,
        )

    def __repr__(self):
        return f"{type(self).__name__}(max_span_attributes={self.max_span_attributes}, max_events_attributes={self.max_event_attributes}, max_link_attributes={self.max_link_attributes}, max_attributes={self.max_attributes}, max_events={self.max_events}, max_links={self.max_links}, max_attribute_length={self.max_attribute_length})"

    @classmethod
    def _from_env_if_absent(
        cls, value: int | None, env_var: str, default: int | None = None
    ) -> int | None:
        if value == cls.UNSET:
            return None

        err_msg = "{} must be a non-negative integer but got {}"

        # if no value is provided for the limit, try to load it from env
        if value is None:
            # return default value if env var is not set
            if env_var not in environ:
                return default

            str_value = environ.get(env_var, "").strip().lower()
            if str_value == _ENV_VALUE_UNSET:
                return None

            try:
                value = int(str_value)
            except ValueError:
                raise ValueError(err_msg.format(env_var, str_value))

        if value < 0:
            raise ValueError(err_msg.format(env_var, value))
        return value


_UnsetLimits = SpanLimits(
    max_attributes=SpanLimits.UNSET,
    max_events=SpanLimits.UNSET,
    max_links=SpanLimits.UNSET,
    max_span_attributes=SpanLimits.UNSET,
    max_event_attributes=SpanLimits.UNSET,
    max_link_attributes=SpanLimits.UNSET,
    max_attribute_length=SpanLimits.UNSET,
    max_span_attribute_length=SpanLimits.UNSET,
)

# not removed for backward compat. please use SpanLimits instead.
SPAN_ATTRIBUTE_COUNT_LIMIT = SpanLimits._from_env_if_absent(  # pylint: disable=protected-access
    None,
    OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT,
    _DEFAULT_OTEL_SPAN_ATTRIBUTE_COUNT_LIMIT,
)


class Span(trace_api.Span, ReadableSpan):
    """See `opentelemetry.trace.Span`.

    Users should create `Span` objects via the `Tracer` instead of this
    constructor.

    Args:
        name: The name of the operation this span represents
        context: The immutable span context
        parent: This span's parent's `opentelemetry.trace.SpanContext`, or
            None if this is a root span
        sampler: The sampler used to create this span
        trace_config:  Unused. Originally intended for trace-level configuration
            from the OpenTelemetry protocol, but the upstream ``TraceConfig``
            proto was removed. Retained for backwards compatibility.
        resource: Entity producing telemetry
        attributes: The span's attributes to be exported
        events: Timestamped events to be exported
        links: Links to other spans to be exported
        span_processor: `SpanProcessor` to invoke when starting and ending
            this `Span`.
        limits: `SpanLimits` instance that was passed to the `TracerProvider`
    """

    def __new__(cls, *args, **kwargs):
        if cls is Span:
            raise TypeError("Span must be instantiated via a tracer.")
        return super().__new__(cls)

    # pylint: disable=too-many-locals
    def __init__(
        self,
        name: str,
        context: trace_api.SpanContext,
        parent: trace_api.SpanContext | None = None,
        sampler: sampling.Sampler | None = None,
        trace_config: None = None,  # TODO
        resource: Resource | None = None,
        attributes: types.Attributes = None,
        events: Sequence[Event] | None = None,
        links: Sequence[trace_api.Link] = (),
        kind: trace_api.SpanKind = trace_api.SpanKind.INTERNAL,
        span_processor: SpanProcessor = SpanProcessor(),
        instrumentation_info: InstrumentationInfo | None = None,
        record_exception: bool = True,
        set_status_on_exception: bool = True,
        limits=_UnsetLimits,
        instrumentation_scope: InstrumentationScope | None = None,
        *,
        record_end_metrics: Callable[[], None] | None = None,
    ) -> None:
        if resource is None:
            resource = Resource.create({})
        super().__init__(
            name=name,
            context=context,
            parent=parent,
            kind=kind,
            resource=resource,
            instrumentation_info=instrumentation_info,
            instrumentation_scope=instrumentation_scope,
        )
        self._sampler = sampler
        self._trace_config = trace_config
        self._record_exception = record_exception
        self._set_status_on_exception = set_status_on_exception
        self._span_processor = span_processor
        self._limits = limits
        self._lock = threading.Lock()
        self._attributes = BoundedAttributes(
            self._limits.max_span_attributes,
            attributes,
            immutable=False,
            max_value_len=self._limits.max_span_attribute_length,
        )
        self._events = self._new_events()
        if events:
            for event in events:
                event._attributes = BoundedAttributes(
                    self._limits.max_event_attributes,
                    event.attributes,
                    max_value_len=self._limits.max_attribute_length,
                )
                self._events.append(event)

        self._links = self._new_links(links)

        self._record_end_metrics = record_end_metrics

    def __repr__(self):
        return f'{type(self).__name__}(name="{self._name}", context={self._context})'

    def _new_events(self):
        return BoundedList(self._limits.max_events)

    def _new_links(self, links: Sequence[trace_api.Link]):
        if not links:
            return BoundedList(self._limits.max_links)

        valid_links = []
        for link in links:
            if link and _is_valid_link(link.context, link.attributes):
                # pylint: disable=protected-access
                link._attributes = BoundedAttributes(
                    self._limits.max_link_attributes,
                    link.attributes,
                    max_value_len=self._limits.max_attribute_length,
                )
                valid_links.append(link)

        return BoundedList.from_seq(self._limits.max_links, valid_links)

    def get_span_context(self) -> trace_api.SpanContext:
        return typing.cast(trace_api.SpanContext, self._context)

    def set_attributes(
        self, attributes: Mapping[str, types.AttributeValue]
    ) -> None:
        with self._lock:
            if self._end_time is not None:
                logger.warning("Setting attribute on ended span.")
                return

            self._attributes._set_items(attributes)  # pylint: disable=protected-access

    def set_attribute(self, key: str, value: types.AttributeValue) -> None:
        with self._lock:
            if self._end_time is not None:
                logger.warning("Setting attribute on ended span.")
                return

            self._attributes[key] = value

    @_check_span_ended
    def _add_event(self, event: EventBase) -> None:
        self._events.append(event)

    def add_event(
        self,
        name: str,
        attributes: types.Attributes = None,
        timestamp: int | None = None,
    ) -> None:
        attributes = BoundedAttributes(
            self._limits.max_event_attributes,
            attributes,
            max_value_len=self._limits.max_attribute_length,
        )
        self._add_event(
            Event(
                name=name,
                attributes=attributes,
                timestamp=timestamp,
            )
        )

    @_check_span_ended
    def _add_link(self, link: trace_api.Link) -> None:
        self._links.append(link)

    def add_link(
        self,
        context: SpanContext,
        attributes: types.Attributes = None,
    ) -> None:
        if not _is_valid_link(context, attributes):
            return

        attributes = BoundedAttributes(
            self._limits.max_link_attributes,
            attributes,
            max_value_len=self._limits.max_attribute_length,
        )
        self._add_link(
            trace_api.Link(
                context=context,
                attributes=attributes,
            )
        )

    def _readable_span(self) -> ReadableSpan:
        return ReadableSpan(
            name=self._name,
            context=self._context,
            parent=self._parent,
            resource=self._resource,
            attributes=self._attributes,
            events=self._events,
            links=self._links,
            kind=self.kind,
            status=self._status,
            start_time=self._start_time,
            end_time=self._end_time,
            instrumentation_info=self._instrumentation_info,
            instrumentation_scope=self._instrumentation_scope,
        )

    def start(
        self,
        start_time: int | None = None,
        parent_context: context_api.Context | None = None,
    ) -> None:
        with self._lock:
            if self._start_time is not None:
                logger.warning("Calling start() on a started span.")
                return
            self._start_time = (
                start_time if start_time is not None else time_ns()
            )

        self._span_processor.on_start(self, parent_context=parent_context)

    def end(self, end_time: int | None = None) -> None:
        with self._lock:
            if self._start_time is None:
                raise RuntimeError("Calling end() on a not started span.")
            if self._end_time is not None:
                logger.warning("Calling end() on an ended span.")
                return

            self._end_time = end_time if end_time is not None else time_ns()
            self._attributes._immutable = True  # pylint: disable=protected-access

        if self._record_end_metrics:
            self._record_end_metrics()
        # pylint: disable=protected-access
        self._span_processor._on_ending(self)
        self._span_processor.on_end(self._readable_span())

    @_check_span_ended
    def update_name(self, name: str) -> None:
        self._name = name

    def is_recording(self) -> bool:
        return self._end_time is None

    @_check_span_ended
    def set_status(
        self,
        status: Status | StatusCode,
        description: str | None = None,
    ) -> None:
        # Ignore future calls if status is already set to OK
        # Ignore calls to set to StatusCode.UNSET
        if isinstance(status, Status):
            if (
                self._status
                and self._status.status_code is StatusCode.OK
                or status.status_code is StatusCode.UNSET
            ):
                return
            if description is not None:
                logger.warning(
                    "Description %s ignored. Use either `Status` or `(StatusCode, Description)`",
                    description,
                )
            self._status = status
        elif isinstance(status, StatusCode):
            if (
                self._status
                and self._status.status_code is StatusCode.OK
                or status is StatusCode.UNSET
            ):
                return
            self._status = Status(status, description)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Ends context manager and calls `end` on the `Span`."""
        if exc_val is not None and self.is_recording():
            # Record the exception as an event
            # pylint:disable=protected-access
            if self._record_exception:
                self.record_exception(exception=exc_val, escaped=True)
            # Records status if span is used as context manager
            # i.e. with tracer.start_span() as span:
            if self._set_status_on_exception:
                self.set_status(
                    Status(
                        status_code=StatusCode.ERROR,
                        description=(f"{type(exc_val).__name__}: {exc_val}"),
                    )
                )

        super().__exit__(exc_type, exc_val, exc_tb)

    def record_exception(
        self,
        exception: BaseException,
        attributes: types.Attributes = None,
        timestamp: int | None = None,
        escaped: bool = False,
    ) -> None:
        """Records an exception as a span event."""
        stacktrace = "".join(traceback.format_exception(exception))
        module = type(exception).__module__
        qualname = type(exception).__qualname__
        exception_type = (
            f"{module}.{qualname}"
            if module and module != "builtins"
            else qualname
        )
        _attributes: MutableMapping[str, types.AttributeValue] = {
            EXCEPTION_TYPE: exception_type,
            EXCEPTION_MESSAGE: str(exception),
            EXCEPTION_STACKTRACE: stacktrace,
            EXCEPTION_ESCAPED: str(escaped),
        }
        if attributes:
            _attributes.update(attributes)
        self.add_event(
            name="exception", attributes=_attributes, timestamp=timestamp
        )


class _Span(Span):
    """Protected implementation of `opentelemetry.trace.Span`.

    This constructor exists to prevent the instantiation of the `Span` class
    by other mechanisms than through the `Tracer`.
    """


@dataclass
class _TracerConfig:
    is_enabled: bool

    @classmethod
    def default(cls):
        return cls(is_enabled=True)


class Tracer(trace_api.Tracer):
    """See `opentelemetry.trace.Tracer`."""

    def __init__(
        self,
        sampler: sampling.Sampler,
        resource: Resource,
        span_processor: SynchronousMultiSpanProcessor
        | ConcurrentMultiSpanProcessor,
        id_generator: IdGenerator,
        instrumentation_info: InstrumentationInfo,
        span_limits: SpanLimits,
        instrumentation_scope: InstrumentationScope,
        *,
        meter_provider: metrics_api.MeterProvider | None = None,
        _tracer_config: _TracerConfig | None = None,
    ) -> None:
        self.sampler = sampler
        self.resource = resource
        self.span_processor = span_processor
        self.id_generator = id_generator
        self.instrumentation_info = instrumentation_info
        self._span_limits = span_limits
        self._instrumentation_scope = instrumentation_scope
        self._tracer_config = _tracer_config or _TracerConfig.default()

        meter_provider = meter_provider or metrics_api.get_meter_provider()
        self._tracer_metrics = create_tracer_metrics(
            meter_provider,
            parse_boolean_environment_variable(
                OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED
            ),
        )

    def _set_tracer_config(self, tracer_config: _TracerConfig):
        self._tracer_config = tracer_config

    def _set_resource(self, resource: Resource) -> None:
        self.resource = resource

    def _is_enabled(self) -> bool:
        """If the tracer is not enabled, start_span will create a NonRecordingSpan"""
        return self._tracer_config.is_enabled

    @_agnosticcontextmanager  # pylint: disable=protected-access
    def start_as_current_span(
        self,
        name: str,
        context: context_api.Context | None = None,
        kind: trace_api.SpanKind = trace_api.SpanKind.INTERNAL,
        attributes: types.Attributes = None,
        links: Sequence[trace_api.Link] | None = (),
        start_time: int | None = None,
        record_exception: bool = True,
        set_status_on_exception: bool = True,
        end_on_exit: bool = True,
    ) -> Iterator[trace_api.Span]:
        span = self.start_span(
            name=name,
            context=context,
            kind=kind,
            attributes=attributes,
            links=links,
            start_time=start_time,
            record_exception=record_exception,
            set_status_on_exception=set_status_on_exception,
        )
        with trace_api.use_span(
            span,
            end_on_exit=end_on_exit,
            record_exception=record_exception,
            set_status_on_exception=set_status_on_exception,
        ) as span:
            yield span

    def start_span(  # pylint: disable=too-many-locals
        self,
        name: str,
        context: context_api.Context | None = None,
        kind: trace_api.SpanKind = trace_api.SpanKind.INTERNAL,
        attributes: types.Attributes = None,
        links: Sequence[trace_api.Link] | None = (),
        start_time: int | None = None,
        record_exception: bool = True,
        set_status_on_exception: bool = True,
    ) -> trace_api.Span:
        links = links or ()
        parent_span_context = trace_api.get_current_span(
            context
        ).get_span_context()

        if parent_span_context is not None and not isinstance(
            parent_span_context, trace_api.SpanContext
        ):
            raise TypeError(
                "parent_span_context must be a SpanContext or None."
            )

        if not self._is_enabled():
            return trace_api.NonRecordingSpan(context=parent_span_context)

        # is_valid determines root span
        if parent_span_context is None or not parent_span_context.is_valid:
            parent_span_context = None
            trace_id = self.id_generator.generate_trace_id()
        else:
            trace_id = parent_span_context.trace_id

        # The sampler decides whether to create a real or no-op span at the
        # time of span creation. No-op spans do not record events, and are not
        # exported.
        # The sampler may also add attributes to the newly-created span, e.g.
        # to include information about the sampling result.
        # The sampler may also modify the parent span context's tracestate
        sampling_result = self.sampler.should_sample(
            context, trace_id, name, kind, attributes, links
        )

        trace_flags = (
            trace_api.TraceFlags(trace_api.TraceFlags.SAMPLED)
            if sampling_result.decision.is_sampled()
            else trace_api.TraceFlags(trace_api.TraceFlags.DEFAULT)
        )

        if parent_span_context is None:
            random_trace_id = self.id_generator.is_trace_id_random()
        else:
            random_trace_id = parent_span_context.trace_flags.random_trace_id

        if random_trace_id:
            trace_flags = trace_api.TraceFlags(
                trace_flags | trace_api.TraceFlags.RANDOM_TRACE_ID
            )

        span_context = trace_api.SpanContext(
            trace_id,
            self.id_generator.generate_span_id(),
            is_remote=False,
            trace_flags=trace_flags,
            trace_state=sampling_result.trace_state,
        )

        record_end_metrics = self._tracer_metrics.start_span(
            parent_span_context, sampling_result.decision
        )

        # Only record if is_recording() is true
        if sampling_result.decision.is_recording():
            # pylint:disable=protected-access
            span = _Span(
                name=name,
                context=span_context,
                parent=parent_span_context,
                sampler=self.sampler,
                resource=self.resource,
                attributes=sampling_result.attributes,
                span_processor=self.span_processor,
                kind=kind,
                links=links,
                instrumentation_info=self.instrumentation_info,
                record_exception=record_exception,
                set_status_on_exception=set_status_on_exception,
                limits=self._span_limits,
                instrumentation_scope=self._instrumentation_scope,
                record_end_metrics=record_end_metrics,
            )
            span.start(start_time=start_time, parent_context=context)
        else:
            span = trace_api.NonRecordingSpan(context=span_context)
        return span


_TracerConfiguratorT = Callable[[InstrumentationScope], _TracerConfig]
_RuleBasedTracerConfigurator = RuleBasedConfigurator[_TracerConfig]


def _default_tracer_configurator(
    tracer_scope: InstrumentationScope,
) -> _TracerConfig:
    """Default Tracer Configurator implementation

    In order to update Tracers configs you need to call
    TracerProvider._set_tracer_configurator with a function
    implementing this interface returning a Tracer Config."""
    return _RuleBasedTracerConfigurator(
        rules=[],
        default_config=_TracerConfig.default(),
    )(tracer_scope)


def _disable_tracer_configurator(
    tracer_scope: InstrumentationScope,
) -> _TracerConfig:
    return _RuleBasedTracerConfigurator(
        rules=[],
        default_config=_TracerConfig(is_enabled=False),
    )(tracer_scope)


class TracerProvider(trace_api.TracerProvider):
    """See `opentelemetry.trace.TracerProvider`."""

    def __init__(
        self,
        sampler: sampling.Sampler | None = None,
        resource: Resource | None = None,
        shutdown_on_exit: bool = True,
        active_span_processor: SynchronousMultiSpanProcessor
        | ConcurrentMultiSpanProcessor
        | None = None,
        id_generator: IdGenerator | None = None,
        span_limits: SpanLimits | None = None,
        *,
        meter_provider: metrics_api.MeterProvider | None = None,
        _tracer_configurator: _TracerConfiguratorT | None = None,
    ) -> None:
        self._active_span_processor = (
            active_span_processor or SynchronousMultiSpanProcessor()
        )
        if id_generator is None:
            self.id_generator = RandomIdGenerator()
        else:
            self.id_generator = id_generator
        if resource is None:
            self._resource = Resource.create({})
        else:
            self._resource = resource
        if not sampler:
            sampler = sampling._get_from_env_or_default()
        self.sampler = sampler
        self._span_limits = span_limits or SpanLimits()
        disabled = environ.get(OTEL_SDK_DISABLED, "")
        self._disabled = disabled.lower().strip() == "true"
        self._atexit_handler = None
        self._meter_provider = meter_provider

        if shutdown_on_exit:
            self._atexit_handler = atexit.register(self.shutdown)

        self._tracer_configurator = (
            _tracer_configurator or _default_tracer_configurator
        )
        self._tracers_lock = threading.Lock()
        self._tracers: dict[InstrumentationScope, Tracer] = {}
        if hasattr(os, "register_at_fork"):
            weak_at_fork = weakref.WeakMethod(self._handle_fork)

            def _after_in_child() -> None:
                if at_fork := weak_at_fork():
                    at_fork()

            os.register_at_fork(after_in_child=_after_in_child)

    def _handle_fork(self) -> None:
        self._tracers_lock = threading.Lock()
        self._update_resource(_get_process_dependent_resource())

    def _set_tracer_configurator(
        self, *, tracer_configurator: _TracerConfiguratorT
    ):
        """This is the function used to update the TracerProvider TracerConfigurator

        Setting a new TracerConfigurator for a TracerProvider will update the
        TracerConfig of all Tracers create by this TracerProvider.
        """
        self._tracer_configurator = tracer_configurator
        with self._tracers_lock:
            for instrumentation_scope, tracer in self._tracers.items():
                tracer_config = self._apply_tracer_configurator(
                    instrumentation_scope
                )
                # pylint: disable-next=protected-access
                tracer._set_tracer_config(tracer_config)

    @property
    def resource(self) -> Resource:
        return self._resource

    def _update_resource(self, resource: Resource) -> None:
        with self._tracers_lock:
            self._resource = self._resource.merge(resource)
            for tracer in self._tracers.values():
                tracer._set_resource(self._resource)  # pylint: disable=protected-access

    def _apply_tracer_configurator(
        self, instrumentation_scope: InstrumentationScope
    ):
        try:
            return self._tracer_configurator(instrumentation_scope)
        except Exception:  # pylint: disable=broad-exception-caught
            logger.exception(
                "Failed to create a Tracer Config for %s, using default Tracer config",
                instrumentation_scope,
            )
            return _TracerConfig.default()

    def get_tracer(
        self,
        instrumenting_module_name: str,
        instrumenting_library_version: str | None = None,
        schema_url: str | None = None,
        attributes: types.Attributes | None = None,
    ) -> "trace_api.Tracer":
        if self._disabled:
            return NoOpTracer()
        if not instrumenting_module_name:  # Reject empty strings too.
            instrumenting_module_name = ""
            logger.error("get_tracer called with missing module name.")
        if instrumenting_library_version is None:
            instrumenting_library_version = ""

        filterwarnings(
            "ignore",
            message=(
                r"You should use InstrumentationScope. Deprecated since version 1.11.1."
            ),
            category=DeprecationWarning,
            module="opentelemetry.sdk.trace",
        )

        instrumentation_info = InstrumentationInfo(
            instrumenting_module_name,
            instrumenting_library_version,
            schema_url,
        )

        instrumentation_scope = InstrumentationScope(
            instrumenting_module_name,
            instrumenting_library_version,
            schema_url,
            attributes,
        )

        with self._tracers_lock:
            if instrumentation_scope in self._tracers:
                return self._tracers[instrumentation_scope]

            tracer_config = self._apply_tracer_configurator(
                instrumentation_scope
            )
            tracer = Tracer(
                self.sampler,
                self.resource,
                self._active_span_processor,
                self.id_generator,
                instrumentation_info,
                self._span_limits,
                instrumentation_scope,
                meter_provider=self._meter_provider,
                _tracer_config=tracer_config,
            )
            self._tracers[instrumentation_scope] = tracer

        return tracer

    def add_span_processor(self, span_processor: SpanProcessor) -> None:
        """Registers a new :class:`SpanProcessor` for this `TracerProvider`.

        The span processors are invoked in the same order they are registered.
        """

        # no lock here because add_span_processor is thread safe for both
        # SynchronousMultiSpanProcessor and ConcurrentMultiSpanProcessor.
        self._active_span_processor.add_span_processor(span_processor)

    def shutdown(self) -> None:
        """Shut down the span processors added to the tracer provider."""
        self._active_span_processor.shutdown()
        if self._atexit_handler is not None:
            atexit.unregister(self._atexit_handler)
            self._atexit_handler = None

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        """Requests the active span processor to process all spans that have not
        yet been processed.

        By default force flush is called sequentially on all added span
        processors. This means that span processors further back in the list
        have less time to flush their spans.
        To have span processors flush their spans in parallel it is possible to
        initialize the tracer provider with an instance of
        `ConcurrentMultiSpanProcessor` at the cost of using multiple threads.

        Args:
            timeout_millis: The maximum amount of time to wait for spans to be
                processed.

        Returns:
            False if the timeout is exceeded, True otherwise.
        """
        return self._active_span_processor.force_flush(timeout_millis)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import math
from collections.abc import Iterator

import pytest
from grpc import Compression as GRPCCompression
from inline_snapshot import snapshot

from opentelemetry._logs import Logger, SeverityNumber
from opentelemetry.exporter.otlp.proto.grpc._log_exporter import (
    OTLPLogExporter as GRPCLogExporter,
)
from opentelemetry.exporter.otlp.proto.http import (
    Compression as HTTPCompression,
)
from opentelemetry.exporter.otlp.proto.http._log_exporter import (
    OTLPLogExporter as HTTPLogExporter,
)
from opentelemetry.sdk._logs import LoggerProvider
from opentelemetry.sdk._logs.export import (
    LogRecordExporter,
    SimpleLogRecordProcessor,
)
from opentelemetry.sdk.resources import Resource
from opentelemetry.test._otlp_test_server import OtlpProtoTestServer

from . import CUSTOM_HEADERS, ExporterConfig, _attrs_to_dict

LOG_EXPORTER_CONFIGS: list[ExporterConfig[LogRecordExporter]] = [
    ExporterConfig(
        id="http",
        exporter_class=HTTPLogExporter,
        kwargs={"endpoint": "http://localhost:4318/v1/logs"},
    ),
    ExporterConfig(
        id="http-deflate",
        exporter_class=HTTPLogExporter,
        kwargs={
            "endpoint": "http://localhost:4318/v1/logs",
            "compression": HTTPCompression.Deflate,
        },
    ),
    ExporterConfig(
        id="http-gzip",
        exporter_class=HTTPLogExporter,
        kwargs={
            "endpoint": "http://localhost:4318/v1/logs",
            "compression": HTTPCompression.Gzip,
        },
    ),
    ExporterConfig(
        id="http-headers",
        exporter_class=HTTPLogExporter,
        kwargs={
            "endpoint": "http://localhost:4318/v1/logs",
            "headers": CUSTOM_HEADERS,
        },
    ),
    ExporterConfig(
        id="grpc",
        exporter_class=GRPCLogExporter,
        kwargs={"insecure": True},
    ),
    ExporterConfig(
        id="grpc-gzip",
        exporter_class=GRPCLogExporter,
        kwargs={"insecure": True, "compression": GRPCCompression.Gzip},
    ),
    ExporterConfig(
        id="grpc-headers",
        exporter_class=GRPCLogExporter,
        kwargs={"insecure": True, "headers": CUSTOM_HEADERS},
    ),
]


class TestLogsExporter:
    @pytest.fixture(
        scope="class", params=LOG_EXPORTER_CONFIGS, ids=lambda c: c.id
    )
    def config(self, request) -> ExporterConfig[LogRecordExporter]:
        return request.param

    @pytest.fixture(scope="class")
    def logger_provider(
        self,
        config: ExporterConfig[LogRecordExporter],
        server: OtlpProtoTestServer,
    ) -> Iterator[LoggerProvider]:
        provider = LoggerProvider(
            resource=Resource.create({"service.name": "test-service"}),
        )
        provider.add_log_record_processor(
            SimpleLogRecordProcessor(config.build())
        )
        try:
            yield provider
        finally:
            provider.shutdown()

    @pytest.fixture(scope="class")
    def logger(self, logger_provider: LoggerProvider) -> Logger:
        return logger_provider.get_logger(__name__)

    @pytest.fixture(autouse=True)
    def clear_server(self, server: OtlpProtoTestServer) -> None:
        server.clear()

    def test_log_body(self, logger: Logger, server: OtlpProtoTestServer):
        logger.emit(body="hello world", severity_number=SeverityNumber.INFO)

        recorded = server.get_log_record(timeout=5.0)
        assert recorded.log_record.body.string_value == snapshot("hello world")

    def test_log_severity_number(
        self, logger: Logger, server: OtlpProtoTestServer
    ):
        logger.emit(
            severity_number=SeverityNumber.ERROR, body="error occurred"
        )

        recorded = server.get_log_record(timeout=5.0)
        assert (
            recorded.log_record.severity_number == SeverityNumber.ERROR.value
        )

    def test_log_severity_text(
        self, logger: Logger, server: OtlpProtoTestServer
    ):
        logger.emit(
            severity_number=SeverityNumber.WARN,
            severity_text="WARN",
            body="warning",
        )

        recorded = server.get_log_record(timeout=5.0)
        assert recorded.log_record.severity_text == snapshot("WARN")

    def test_log_attributes(self, logger: Logger, server: OtlpProtoTestServer):
        logger.emit(
            body="attrs test",
            severity_number=SeverityNumber.INFO,
            attributes={
                "str_key": "hello",
                "int_key": 42,
                "float_key": 3.14,
                "bool_key": True,
            },
        )

        recorded = server.get_log_record(timeout=5.0)
        attrs = _attrs_to_dict(recorded.log_record.attributes)
        assert math.isclose(attrs.pop("float_key"), 3.14, abs_tol=1e-5)
        assert attrs == snapshot(
            {"str_key": "hello", "int_key": 42, "bool_key": True}
        )

    def test_scope_attributes(
        self, logger_provider: LoggerProvider, server: OtlpProtoTestServer
    ):
        logger = logger_provider.get_logger(
            "test.scope",
            version="1.0.0",
            attributes={"scope.key": "scope.val"},
        )
        logger.emit(body="scope test", severity_number=SeverityNumber.INFO)

        recorded = server.get_log_record(timeout=5.0)
        assert recorded.scope.name == snapshot("test.scope")
        assert recorded.scope.version == snapshot("1.0.0")
        assert _attrs_to_dict(recorded.scope.attributes) == snapshot(
            {"scope.key": "scope.val"}
        )

    def test_resource_attributes(
        self, logger: Logger, server: OtlpProtoTestServer
    ):
        logger.emit(body="resource test", severity_number=SeverityNumber.INFO)

        recorded = server.get_log_record(timeout=5.0)
        resource_attrs = _attrs_to_dict(recorded.resource.attributes)
        assert resource_attrs["service.name"] == snapshot("test-service")



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=invalid-name,no-self-use

from threading import Event, Thread
from time import sleep
from unittest import TestCase
from unittest.mock import MagicMock, Mock, patch

from opentelemetry.sdk.metrics._internal.measurement_consumer import (
    MeasurementConsumer,
    SynchronousMeasurementConsumer,
)
from opentelemetry.sdk.metrics._internal.sdk_configuration import (
    SdkConfiguration,
)


@patch(
    "opentelemetry.sdk.metrics._internal."
    "measurement_consumer.MetricReaderStorage"
)
class TestSynchronousMeasurementConsumer(TestCase):
    def test_parent(self, _):
        self.assertIsInstance(
            SynchronousMeasurementConsumer(MagicMock(), metric_readers=()),
            MeasurementConsumer,
        )

    def test_creates_metric_reader_storages(self, MockMetricReaderStorage):
        """It should create one MetricReaderStorage per metric reader passed in the SdkConfiguration"""
        reader_mocks = [Mock() for _ in range(5)]
        SynchronousMeasurementConsumer(
            SdkConfiguration(
                exemplar_filter=Mock(),
                resource=Mock(),
                views=Mock(),
            ),
            metric_readers=reader_mocks,
        )
        self.assertEqual(len(MockMetricReaderStorage.mock_calls), 5)

    def test_measurements_passed_to_each_reader_storage(
        self, MockMetricReaderStorage
    ):
        reader_mocks = [Mock() for _ in range(5)]
        reader_storage_mocks = [Mock() for _ in range(5)]
        MockMetricReaderStorage.side_effect = reader_storage_mocks

        consumer = SynchronousMeasurementConsumer(
            SdkConfiguration(
                exemplar_filter=Mock(should_sample=Mock(return_value=False)),
                resource=Mock(),
                views=Mock(),
            ),
            metric_readers=reader_mocks,
        )
        measurement_mock = Mock()
        consumer.consume_measurement(measurement_mock)

        for rs_mock in reader_storage_mocks:
            rs_mock.consume_measurement.assert_called_once_with(
                measurement_mock, False
            )

    def test_collect_passed_to_reader_stage(self, MockMetricReaderStorage):
        """Its collect() method should defer to the underlying MetricReaderStorage"""
        reader_mocks = [Mock() for _ in range(5)]
        reader_storage_mocks = [Mock() for _ in range(5)]
        MockMetricReaderStorage.side_effect = reader_storage_mocks

        consumer = SynchronousMeasurementConsumer(
            SdkConfiguration(
                exemplar_filter=Mock(),
                resource=Mock(),
                views=Mock(),
            ),
            metric_readers=reader_mocks,
        )
        for r_mock, rs_mock in zip(reader_mocks, reader_storage_mocks):
            rs_mock.collect.assert_not_called()
            consumer.collect(r_mock)
            rs_mock.collect.assert_called_once_with()

    def test_collect_calls_async_instruments(self, MockMetricReaderStorage):
        """Its collect() method should invoke async instruments and pass measurements to the
        corresponding metric reader storage"""
        reader_mock = Mock()
        reader_storage_mock = Mock()
        MockMetricReaderStorage.return_value = reader_storage_mock
        consumer = SynchronousMeasurementConsumer(
            SdkConfiguration(
                exemplar_filter=Mock(should_sample=Mock(return_value=False)),
                resource=Mock(),
                views=Mock(),
            ),
            metric_readers=[reader_mock],
        )
        async_instrument_mocks = [MagicMock() for _ in range(5)]
        for i_mock in async_instrument_mocks:
            i_mock.callback.return_value = [Mock()]
            consumer.register_asynchronous_instrument(i_mock)

        consumer.collect(reader_mock)

        # it should call async instruments
        for i_mock in async_instrument_mocks:
            i_mock.callback.assert_called_once()

        # it should pass measurements to reader storage
        self.assertEqual(
            len(reader_storage_mock.consume_measurement.mock_calls), 5
        )
        # assert consume_measurement was called with at least 2 arguments the second
        # matching the mocked exemplar filter
        self.assertFalse(reader_storage_mock.consume_measurement.call_args[1])

    def test_collect_timeout(self, MockMetricReaderStorage):
        reader_mock = Mock()
        reader_storage_mock = Mock()
        MockMetricReaderStorage.return_value = reader_storage_mock
        consumer = SynchronousMeasurementConsumer(
            SdkConfiguration(
                exemplar_filter=Mock(),
                resource=Mock(),
                views=Mock(),
            ),
            metric_readers=[reader_mock],
        )

        def sleep_1(*args, **kwargs):
            sleep(1)

        consumer.register_asynchronous_instrument(
            Mock(**{"callback.side_effect": sleep_1})
        )

        with self.assertRaises(Exception) as error:
            consumer.collect(reader_mock, timeout_millis=10)

        self.assertIn(
            "Timed out while executing callback", error.exception.args[0]
        )

    @patch(
        "opentelemetry.sdk.metrics._internal."
        "measurement_consumer.CallbackOptions"
    )
    @patch("opentelemetry.sdk.metrics._internal.measurement_consumer.time_ns")
    def test_collect_deadline(
        self, mock_time_ns, mock_callback_options, MockMetricReaderStorage
    ):
        reader_mock = Mock()
        reader_storage_mock = Mock()
        MockMetricReaderStorage.return_value = reader_storage_mock
        consumer = SynchronousMeasurementConsumer(
            SdkConfiguration(
                exemplar_filter=Mock(),
                resource=Mock(),
                views=Mock(),
            ),
            metric_readers=[reader_mock],
        )

        consumer.register_asynchronous_instrument(
            Mock(**{"callback.return_value": []})
        )
        consumer.register_asynchronous_instrument(
            Mock(**{"callback.return_value": []})
        )

        # collect start, first remaining_time, post-first callback,
        # second remaining_time, post-second callback
        mock_time_ns.side_effect = [
            0,
            0,
            int(1e9),
            int(1e9),
            int(2e9),
        ]

        consumer.collect(reader_mock)

        callback_options_time_call = mock_callback_options.mock_calls[
            -1
        ].kwargs["timeout_millis"]

        self.assertLess(
            callback_options_time_call,
            10000,
        )


class TestSynchronousMeasurementConsumerConcurrency(TestCase):
    def test_consume_measurement_does_not_acquire_lock(self):
        """consume_measurement must stay lock free on the hot path."""
        with patch(
            "opentelemetry.sdk.metrics._internal."
            "measurement_consumer.MetricReaderStorage"
        ):
            consumer = SynchronousMeasurementConsumer(
                SdkConfiguration(
                    exemplar_filter=Mock(
                        should_sample=Mock(return_value=False)
                    ),
                    resource=Mock(),
                    views=Mock(),
                ),
                metric_readers=[Mock()],
            )

        mock_lock = MagicMock()
        # pylint: disable-next=protected-access
        consumer._lock = mock_lock

        consumer.consume_measurement(Mock())

        mock_lock.__enter__.assert_not_called()

    def test_concurrent_changes_to_metric_readers(self):
        timeout = 1
        failure = None
        iteration_started = Event()
        mutation_done = Event()
        iteration_timeout_error = "Timed out waiting for iteration to start"
        mutation_timeout_error = "Timed out waiting for mutation to be done"

        consumer = SynchronousMeasurementConsumer(
            SdkConfiguration(
                exemplar_filter=MagicMock(),
                resource=MagicMock(),
                views=MagicMock(),
            ),
            metric_readers=[MagicMock()],
        )

        def _hooked_iter(iterable):
            nonlocal failure

            iterable = iter(iterable)
            iteration_started.set()
            if not mutation_done.wait(timeout):
                failure = mutation_timeout_error
            yield next(iterable, None)
            yield from iterable

        class HookedDict(dict):
            def values(self):
                return _hooked_iter(super().values())

        with patch.object(
            consumer,
            "_reader_storages",
            # pylint: disable-next=protected-access
            HookedDict(consumer._reader_storages),
        ):

            def mutate():
                """Directly mutate _reader_storages after iteration starts"""
                nonlocal failure
                if not iteration_started.wait(timeout):
                    failure = iteration_timeout_error
                # pylint: disable-next=protected-access
                consumer._reader_storages.clear()
                mutation_done.set()

            # Verify that test setup works (direct mutation with no synchronization fails)
            with self.assertRaises(RuntimeError) as cm:
                t = Thread(target=mutate)
                t.start()
                try:
                    consumer.consume_measurement(MagicMock())
                finally:
                    t.join()
            self.assertEqual(
                "dictionary changed size during iteration", str(cm.exception)
            )
            self.assertIsNone(failure)

        # Reset the events for the second scenario
        iteration_started.clear()
        mutation_done.clear()
        failure = None

        with patch.object(
            consumer,
            "_reader_storages",
            # pylint: disable-next=protected-access
            HookedDict(consumer._reader_storages),
        ):

            def add_and_remove_readers():
                """Mutate via the public API while consume_measurement runs"""
                nonlocal failure
                if not iteration_started.wait(timeout):
                    failure = iteration_timeout_error
                reader = MagicMock()
                consumer.add_metric_reader(reader)
                consumer.remove_metric_reader(reader)
                mutation_done.set()

            # The copy-on-write API never mutates the mapping being
            # iterated, so `consume_measurement` completes without raising even
            # though add/remove run concurrently mid-iteration.
            t = Thread(target=add_and_remove_readers)
            t.start()
            try:
                consumer.consume_measurement(MagicMock())
            finally:
                t.join()
            self.assertIsNone(failure)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import asyncio

from opentracing.ext import tags

# pylint: disable=import-error
from ..otel_ot_shim_tracer import MockTracer
from ..testcase import OpenTelemetryTestCase
from ..utils import get_one_by_tag
from .response_listener import ResponseListener


async def task(message, listener):
    res = f"{message}::response"
    listener.on_response(res)
    return res


class Client:
    def __init__(self, tracer, loop):
        self.tracer = tracer
        self.loop = loop

    def send_sync(self, message):
        span = self.tracer.start_span("send")
        span.set_tag(tags.SPAN_KIND, tags.SPAN_KIND_RPC_CLIENT)

        listener = ResponseListener(span)
        return self.loop.run_until_complete(task(message, listener))


class TestAsyncio(OpenTelemetryTestCase):
    def setUp(self):
        self.tracer = MockTracer()
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def tearDown(self):
        self.loop.close()
        super().tearDown()

    def test_main(self):
        client = Client(self.tracer, self.loop)
        res = client.send_sync("message")
        self.assertEqual(res, "message::response")

        spans = self.tracer.finished_spans()
        self.assertEqual(len(spans), 1)

        span = get_one_by_tag(spans, tags.SPAN_KIND, tags.SPAN_KIND_RPC_CLIENT)
        self.assertIsNotNone(span)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from opentelemetry.test import TestCase


class TestAssertNotRaises(TestCase):
    def test_no_exception(self):
        try:
            with self.assertNotRaises(Exception):
                pass

        except Exception as error:  # pylint: disable=broad-exception-caught
            self.fail(  # pylint: disable=no-member
                f"Unexpected exception {error} was raised"
            )

    def test_no_specified_exception_single(self):
        try:
            with self.assertNotRaises(KeyError):
                1 / 0  # pylint: disable=pointless-statement

        except Exception as error:  # pylint: disable=broad-exception-caught
            self.fail(  # pylint: disable=no-member
                f"Unexpected exception {error} was raised"
            )

    def test_no_specified_exception_multiple(self):
        try:
            with self.assertNotRaises(KeyError, IndexError):
                1 / 0  # pylint: disable=pointless-statement

        except Exception as error:  # pylint: disable=broad-exception-caught
            self.fail(  # pylint: disable=no-member
                f"Unexpected exception {error} was raised"
            )

    def test_exception(self):
        with self.assertRaises(AssertionError):
            with self.assertNotRaises(ZeroDivisionError):
                1 / 0  # pylint: disable=pointless-statement

    def test_missing_exception(self):
        with self.assertRaises(AssertionError) as error:
            with self.assertNotRaises(ZeroDivisionError):

                def raise_zero_division_error():
                    raise ZeroDivisionError()

                raise_zero_division_error()

        error_lines = error.exception.args[0].split("\n")
        stripped_error_lines = [line.strip() for line in error_lines]

        self.assertIn("Unexpected exception was raised:", stripped_error_lines)
        self.assertIn("raise_zero_division_error()", stripped_error_lines)
        self.assertIn("raise ZeroDivisionError()", stripped_error_lines)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=protected-access,too-many-lines,invalid-name
# pylint: disable=consider-using-enumerate,no-self-use,too-many-public-methods

from inspect import currentframe
from itertools import permutations
from logging import WARNING
from math import inf, ldexp
from random import Random, randrange
from sys import float_info, maxsize
from time import time_ns
from types import MethodType
from unittest.mock import Mock, patch

from opentelemetry.context import Context
from opentelemetry.sdk.metrics._internal.aggregation import (
    AggregationTemporality,
    _ExponentialBucketHistogramAggregation,
)
from opentelemetry.sdk.metrics._internal.exponential_histogram.buckets import (
    Buckets,
)
from opentelemetry.sdk.metrics._internal.exponential_histogram.mapping.exponent_mapping import (
    ExponentMapping,
)
from opentelemetry.sdk.metrics._internal.exponential_histogram.mapping.ieee_754 import (
    MAX_NORMAL_EXPONENT,
    MIN_NORMAL_EXPONENT,
)
from opentelemetry.sdk.metrics._internal.exponential_histogram.mapping.logarithm_mapping import (
    LogarithmMapping,
)
from opentelemetry.sdk.metrics._internal.measurement import Measurement
from opentelemetry.sdk.metrics._internal.point import (
    ExponentialHistogramDataPoint,
)
from opentelemetry.sdk.metrics._internal.view import _default_reservoir_factory
from opentelemetry.sdk.metrics.view import (
    ExponentialBucketHistogramAggregation,
)
from opentelemetry.test import TestCase


def get_counts(buckets: Buckets) -> int:
    counts = []

    for index in range(len(buckets)):
        counts.append(buckets[index])

    return counts


def center_val(mapping: ExponentMapping, index: int) -> float:
    return (
        mapping.get_lower_boundary(index)
        + mapping.get_lower_boundary(index + 1)
    ) / 2


def swap(
    first: _ExponentialBucketHistogramAggregation,
    second: _ExponentialBucketHistogramAggregation,
):
    for attribute in [
        "_value_positive",
        "_value_negative",
        "_sum",
        "_count",
        "_zero_count",
        "_min",
        "_max",
        "_mapping",
    ]:
        temp = getattr(first, attribute)
        setattr(first, attribute, getattr(second, attribute))
        setattr(second, attribute, temp)


class TestExponentialBucketHistogramAggregation(TestCase):
    @patch("opentelemetry.sdk.metrics._internal.aggregation.LogarithmMapping")
    def test_create_aggregation(self, mock_logarithm_mapping):
        exponential_bucket_histogram_aggregation = (
            ExponentialBucketHistogramAggregation()
        )._create_aggregation(Mock(), Mock(), Mock(), Mock())

        self.assertEqual(
            exponential_bucket_histogram_aggregation._max_scale, 20
        )

        mock_logarithm_mapping.assert_called_with(20)

        exponential_bucket_histogram_aggregation = (
            ExponentialBucketHistogramAggregation(max_scale=10)
        )._create_aggregation(Mock(), Mock(), Mock(), Mock())

        self.assertEqual(
            exponential_bucket_histogram_aggregation._max_scale, 10
        )

        mock_logarithm_mapping.assert_called_with(10)

        with self.assertLogs(level=WARNING):
            exponential_bucket_histogram_aggregation = (
                ExponentialBucketHistogramAggregation(max_scale=100)
            )._create_aggregation(Mock(), Mock(), Mock(), Mock())

        self.assertEqual(
            exponential_bucket_histogram_aggregation._max_scale, 100
        )

        mock_logarithm_mapping.assert_called_with(100)

    def test_create_aggregation_record_min_max(self):
        for record_min_max, expected in [
            (None, True),
            (True, True),
            (False, False),
        ]:
            with self.subTest(record_min_max=record_min_max):
                kwargs = (
                    {}
                    if record_min_max is None
                    else {"record_min_max": record_min_max}
                )
                exponential_bucket_histogram_aggregation = (
                    ExponentialBucketHistogramAggregation(**kwargs)
                )._create_aggregation(Mock(), Mock(), Mock(), Mock())

                self.assertEqual(
                    exponential_bucket_histogram_aggregation._record_min_max,
                    expected,
                )

    def test_min_max(self):
        """
        `record_min_max` indicates the aggregator to record the minimum and
        maximum value in the population
        """

        now = time_ns()

        for record_min_max, expected_min, expected_max in [
            (True, 1, 9999),
            (False, inf, -inf),
        ]:
            with self.subTest(record_min_max=record_min_max):
                ctx = Context()
                exponential_histogram_aggregation = (
                    _ExponentialBucketHistogramAggregation(
                        Mock(),
                        _default_reservoir_factory(
                            _ExponentialBucketHistogramAggregation
                        ),
                        AggregationTemporality.CUMULATIVE,
                        0,
                        record_min_max=record_min_max,
                    )
                )

                for value in [2, 4, 1, 9999]:
                    exponential_histogram_aggregation.aggregate(
                        Measurement(value, now, Mock(), ctx)
                    )

                self.assertEqual(
                    exponential_histogram_aggregation._min, expected_min
                )
                self.assertEqual(
                    exponential_histogram_aggregation._max, expected_max
                )

    def assertInEpsilon(self, first, second, epsilon):
        self.assertLessEqual(first, (second * (1 + epsilon)))
        self.assertGreaterEqual(first, (second * (1 - epsilon)))

    def require_equal(self, a, b):
        if a._sum == 0 or b._sum == 0:
            self.assertAlmostEqual(a._sum, b._sum, 1e-6)
        else:
            self.assertInEpsilon(a._sum, b._sum, 1e-6)

        self.assertEqual(a._count, b._count)
        self.assertEqual(a._zero_count, b._zero_count)

        self.assertEqual(a._mapping.scale, b._mapping.scale)

        self.assertEqual(len(a._value_positive), len(b._value_positive))
        self.assertEqual(len(a._value_negative), len(b._value_negative))

        for index in range(len(a._value_positive)):
            self.assertEqual(
                a._value_positive[index], b._value_positive[index]
            )

        for index in range(len(a._value_negative)):
            self.assertEqual(
                a._value_negative[index], b._value_negative[index]
            )

    def test_alternating_growth_0(self):
        """
        Tests insertion of [2, 4, 1].  The index of 2 (i.e., 0) becomes
        `indexBase`, the 4 goes to its right and the 1 goes in the last
        position of the backing array.  With 3 binary orders of magnitude
        and MaxSize=4, this must finish with scale=0; with minimum value 1
        this must finish with offset=-1 (all scales).

        """

        # The corresponding Go test is TestAlternatingGrowth1 where:
        # agg := NewFloat64(NewConfig(WithMaxSize(4)))
        # agg is an instance of github.com/lightstep/otel-launcher-go/lightstep/sdk/metric/aggregator/histogram/structure.Histogram[float64]

        exponential_histogram_aggregation = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
                max_size=4,
            )
        )

        now = time_ns()
        ctx = Context()
        exponential_histogram_aggregation.aggregate(
            Measurement(2, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(4, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(1, now, Mock(), ctx)
        )

        self.assertEqual(
            exponential_histogram_aggregation._value_positive.offset, -1
        )
        self.assertEqual(exponential_histogram_aggregation._mapping.scale, 0)
        self.assertEqual(
            get_counts(exponential_histogram_aggregation._value_positive),
            [1, 1, 1],
        )

    def test_alternating_growth_1(self):
        """
        Tests insertion of [2, 2, 4, 1, 8, 0.5].  The test proceeds as¶
        above but then downscales once further to scale=-1, thus index -1¶
        holds range [0.25, 1.0), index 0 holds range [1.0, 4), index 1¶
        holds range [4, 16).¶
        """

        exponential_histogram_aggregation = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
                max_size=4,
            )
        )

        now = time_ns()
        ctx = Context()
        exponential_histogram_aggregation.aggregate(
            Measurement(2, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(2, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(2, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(1, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(8, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(0.5, now, Mock(), ctx)
        )

        self.assertEqual(
            exponential_histogram_aggregation._value_positive.offset, -1
        )
        self.assertEqual(exponential_histogram_aggregation._mapping.scale, -1)
        self.assertEqual(
            get_counts(exponential_histogram_aggregation._value_positive),
            [2, 3, 1],
        )

    def test_permutations(self):
        """
        Tests that every permutation of certain sequences with maxSize=2
        results in the same scale=-1 histogram.
        """

        now = time_ns()
        ctx = Context()
        for test_values, expected in [
            [
                [0.5, 1.0, 2.0],
                {
                    "scale": -1,
                    "offset": -1,
                    "len": 2,
                    "at_0": 2,
                    "at_1": 1,
                },
            ],
            [
                [1.0, 2.0, 4.0],
                {
                    "scale": -1,
                    "offset": -1,
                    "len": 2,
                    "at_0": 1,
                    "at_1": 2,
                },
            ],
            [
                [0.25, 0.5, 1],
                {
                    "scale": -1,
                    "offset": -2,
                    "len": 2,
                    "at_0": 1,
                    "at_1": 2,
                },
            ],
        ]:
            for permutation in permutations(test_values):
                exponential_histogram_aggregation = (
                    _ExponentialBucketHistogramAggregation(
                        Mock(),
                        _default_reservoir_factory(
                            _ExponentialBucketHistogramAggregation
                        ),
                        AggregationTemporality.DELTA,
                        Mock(),
                        max_size=2,
                    )
                )

                for value in permutation:
                    exponential_histogram_aggregation.aggregate(
                        Measurement(value, now, Mock(), ctx)
                    )

                self.assertEqual(
                    exponential_histogram_aggregation._mapping.scale,
                    expected["scale"],
                )
                self.assertEqual(
                    exponential_histogram_aggregation._value_positive.offset,
                    expected["offset"],
                )
                self.assertEqual(
                    len(exponential_histogram_aggregation._value_positive),
                    expected["len"],
                )
                self.assertEqual(
                    exponential_histogram_aggregation._value_positive[0],
                    expected["at_0"],
                )
                self.assertEqual(
                    exponential_histogram_aggregation._value_positive[1],
                    expected["at_1"],
                )

    def test_ascending_sequence(self):
        for max_size in [3, 4, 6, 9]:
            for offset in range(-5, 6):
                for init_scale in [0, 4]:
                    self.ascending_sequence_test(max_size, offset, init_scale)

    # pylint: disable=too-many-locals
    def ascending_sequence_test(
        self, max_size: int, offset: int, init_scale: int
    ):
        now = time_ns()
        ctx = Context()
        for step in range(max_size, max_size * 4):
            exponential_histogram_aggregation = (
                _ExponentialBucketHistogramAggregation(
                    Mock(),
                    _default_reservoir_factory(
                        _ExponentialBucketHistogramAggregation
                    ),
                    AggregationTemporality.DELTA,
                    Mock(),
                    max_size=max_size,
                )
            )

            if init_scale <= 0:
                mapping = ExponentMapping(init_scale)
            else:
                mapping = LogarithmMapping(init_scale)

            min_val = center_val(mapping, offset)
            max_val = center_val(mapping, offset + step)

            sum_ = 0.0

            for index in range(max_size):
                value = center_val(mapping, offset + index)
                exponential_histogram_aggregation.aggregate(
                    Measurement(value, now, Mock(), ctx)
                )
                sum_ += value

            self.assertEqual(
                init_scale, exponential_histogram_aggregation._mapping._scale
            )
            self.assertEqual(
                offset,
                exponential_histogram_aggregation._value_positive.offset,
            )

            exponential_histogram_aggregation.aggregate(
                Measurement(max_val, now, Mock(), ctx)
            )
            sum_ += max_val

            self.assertNotEqual(
                0, exponential_histogram_aggregation._value_positive[0]
            )

            # The maximum-index filled bucket is at or
            # above the mid-point, (otherwise we
            # downscaled too much).

            max_fill = 0
            total_count = 0

            for index in range(
                len(exponential_histogram_aggregation._value_positive)
            ):
                total_count += (
                    exponential_histogram_aggregation._value_positive[index]
                )
                if (
                    exponential_histogram_aggregation._value_positive[index]
                    != 0
                ):
                    max_fill = index

            # FIXME the corresponding Go code is
            # require.GreaterOrEqual(t, maxFill, uint32(maxSize)/2), make sure
            # this is actually equivalent.
            self.assertGreaterEqual(max_fill, int(max_size / 2))

            self.assertGreaterEqual(max_size + 1, total_count)
            self.assertGreaterEqual(
                max_size + 1, exponential_histogram_aggregation._count
            )
            self.assertGreaterEqual(
                sum_, exponential_histogram_aggregation._sum
            )

            if init_scale <= 0:
                mapping = ExponentMapping(
                    exponential_histogram_aggregation._mapping.scale
                )
            else:
                mapping = LogarithmMapping(
                    exponential_histogram_aggregation._mapping.scale
                )
            index = mapping.map_to_index(min_val)

            self.assertEqual(
                index, exponential_histogram_aggregation._value_positive.offset
            )

            index = mapping.map_to_index(max_val)

            self.assertEqual(
                index,
                exponential_histogram_aggregation._value_positive.offset
                + len(exponential_histogram_aggregation._value_positive)
                - 1,
            )

    def test_reset(self):
        now = time_ns()
        ctx = Context()
        for increment in [0x1, 0x100, 0x10000, 0x100000000, 0x200000000]:

            def mock_increment(self, bucket_index: int) -> None:
                """
                Increments a bucket
                """
                # pylint: disable=cell-var-from-loop
                self._counts[bucket_index] += increment

            exponential_histogram_aggregation = (
                _ExponentialBucketHistogramAggregation(
                    Mock(),
                    _default_reservoir_factory(
                        _ExponentialBucketHistogramAggregation
                    ),
                    AggregationTemporality.DELTA,
                    Mock(),
                    max_size=256,
                )
            )

            self.assertEqual(
                exponential_histogram_aggregation._count,
                exponential_histogram_aggregation._zero_count,
            )
            self.assertEqual(0, exponential_histogram_aggregation._sum)
            expect = 0

            exponential_histogram_aggregation._value_positive = Buckets()

            for value in range(2, 257):
                expect += value * increment
                with patch.object(
                    exponential_histogram_aggregation._value_positive,
                    "increment_bucket",
                    MethodType(
                        mock_increment,
                        exponential_histogram_aggregation._value_positive,
                    ),
                ):
                    exponential_histogram_aggregation.aggregate(
                        Measurement(value, now, Mock(), ctx)
                    )
            exponential_histogram_aggregation._count *= increment
            exponential_histogram_aggregation._sum *= increment

            self.assertEqual(expect, exponential_histogram_aggregation._sum)
            self.assertEqual(
                255 * increment, exponential_histogram_aggregation._count
            )

            # See test_integer_aggregation about why scale is 5, len is
            # 256 - (1 << scale)- 1 and offset is (1 << scale) - 1.
            scale = exponential_histogram_aggregation._mapping.scale
            self.assertEqual(5, scale)

            self.assertEqual(
                256 - ((1 << scale) - 1),
                len(exponential_histogram_aggregation._value_positive),
            )
            self.assertEqual(
                (1 << scale) - 1,
                exponential_histogram_aggregation._value_positive.offset,
            )

            for index in range(0, 256):
                self.assertLessEqual(
                    exponential_histogram_aggregation._value_positive[index],
                    6 * increment,
                )

    def test_move_into(self):
        now = time_ns()
        ctx = Context()

        exponential_histogram_aggregation_0 = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
                max_size=256,
            )
        )
        exponential_histogram_aggregation_1 = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
                max_size=256,
            )
        )

        expect = 0

        for index in range(2, 257):
            expect += index
            exponential_histogram_aggregation_0.aggregate(
                Measurement(index, now, Mock(), ctx)
            )
            exponential_histogram_aggregation_0.aggregate(
                Measurement(0, now, Mock(), ctx)
            )

        swap(
            exponential_histogram_aggregation_0,
            exponential_histogram_aggregation_1,
        )

        self.assertEqual(0, exponential_histogram_aggregation_0._sum)
        self.assertEqual(0, exponential_histogram_aggregation_0._count)
        self.assertEqual(0, exponential_histogram_aggregation_0._zero_count)

        self.assertEqual(expect, exponential_histogram_aggregation_1._sum)
        self.assertEqual(255 * 2, exponential_histogram_aggregation_1._count)
        self.assertEqual(255, exponential_histogram_aggregation_1._zero_count)

        scale = exponential_histogram_aggregation_1._mapping.scale
        self.assertEqual(5, scale)

        self.assertEqual(
            256 - ((1 << scale) - 1),
            len(exponential_histogram_aggregation_1._value_positive),
        )
        self.assertEqual(
            (1 << scale) - 1,
            exponential_histogram_aggregation_1._value_positive.offset,
        )

        for index in range(0, 256):
            self.assertLessEqual(
                exponential_histogram_aggregation_1._value_positive[index], 6
            )

    def test_very_large_numbers(self):
        now = time_ns()
        ctx = Context()

        exponential_histogram_aggregation = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
                max_size=2,
            )
        )

        def expect_balanced(count: int):
            self.assertEqual(
                2, len(exponential_histogram_aggregation._value_positive)
            )
            self.assertEqual(
                -1, exponential_histogram_aggregation._value_positive.offset
            )
            self.assertEqual(
                count, exponential_histogram_aggregation._value_positive[0]
            )
            self.assertEqual(
                count, exponential_histogram_aggregation._value_positive[1]
            )

        exponential_histogram_aggregation.aggregate(
            Measurement(2**-100, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(2**100, now, Mock(), ctx)
        )

        self.assertLessEqual(
            2**100, (exponential_histogram_aggregation._sum * (1 + 1e-5))
        )
        self.assertGreaterEqual(
            2**100, (exponential_histogram_aggregation._sum * (1 - 1e-5))
        )

        self.assertEqual(2, exponential_histogram_aggregation._count)
        self.assertEqual(-7, exponential_histogram_aggregation._mapping.scale)

        expect_balanced(1)

        exponential_histogram_aggregation.aggregate(
            Measurement(2**-127, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(2**128, now, Mock(), ctx)
        )

        self.assertLessEqual(
            2**128, (exponential_histogram_aggregation._sum * (1 + 1e-5))
        )
        self.assertGreaterEqual(
            2**128, (exponential_histogram_aggregation._sum * (1 - 1e-5))
        )

        self.assertEqual(4, exponential_histogram_aggregation._count)
        self.assertEqual(-7, exponential_histogram_aggregation._mapping.scale)

        expect_balanced(2)

        exponential_histogram_aggregation.aggregate(
            Measurement(2**-129, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(2**255, now, Mock(), ctx)
        )

        self.assertLessEqual(
            2**255, (exponential_histogram_aggregation._sum * (1 + 1e-5))
        )
        self.assertGreaterEqual(
            2**255, (exponential_histogram_aggregation._sum * (1 - 1e-5))
        )
        self.assertEqual(6, exponential_histogram_aggregation._count)
        self.assertEqual(-8, exponential_histogram_aggregation._mapping.scale)

        expect_balanced(3)

    def test_full_range(self):
        now = time_ns()
        ctx = Context()

        exponential_histogram_aggregation = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
                max_size=2,
            )
        )

        exponential_histogram_aggregation.aggregate(
            Measurement(float_info.max, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(1, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(2**-1074, now, Mock(), ctx)
        )

        self.assertEqual(
            float_info.max, exponential_histogram_aggregation._sum
        )
        self.assertEqual(3, exponential_histogram_aggregation._count)
        self.assertEqual(
            ExponentMapping._min_scale,
            exponential_histogram_aggregation._mapping.scale,
        )

        self.assertEqual(
            _ExponentialBucketHistogramAggregation._min_max_size,
            len(exponential_histogram_aggregation._value_positive),
        )
        self.assertEqual(
            -1, exponential_histogram_aggregation._value_positive.offset
        )
        self.assertLessEqual(
            exponential_histogram_aggregation._value_positive[0], 2
        )
        self.assertLessEqual(
            exponential_histogram_aggregation._value_positive[1], 1
        )

    def test_aggregator_min_max(self):
        now = time_ns()
        ctx = Context()
        exponential_histogram_aggregation = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
            )
        )

        for value in [1, 3, 5, 7, 9]:
            exponential_histogram_aggregation.aggregate(
                Measurement(value, now, Mock(), ctx)
            )

        self.assertEqual(1, exponential_histogram_aggregation._min)
        self.assertEqual(9, exponential_histogram_aggregation._max)

        exponential_histogram_aggregation = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
            )
        )

        for value in [-1, -3, -5, -7, -9]:
            exponential_histogram_aggregation.aggregate(
                Measurement(value, now, Mock(), ctx)
            )

        self.assertEqual(-9, exponential_histogram_aggregation._min)
        self.assertEqual(-1, exponential_histogram_aggregation._max)

    def test_aggregator_copy_swap(self):
        now = time_ns()
        ctx = Context()
        exponential_histogram_aggregation_0 = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
            )
        )
        for value in [1, 3, 5, 7, 9, -1, -3, -5]:
            exponential_histogram_aggregation_0.aggregate(
                Measurement(value, now, Mock(), ctx)
            )
        exponential_histogram_aggregation_1 = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
            )
        )
        for value in [5, 4, 3, 2]:
            exponential_histogram_aggregation_1.aggregate(
                Measurement(value, now, Mock(), ctx)
            )
        exponential_histogram_aggregation_2 = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
            )
        )

        swap(
            exponential_histogram_aggregation_0,
            exponential_histogram_aggregation_1,
        )

        # pylint: disable=unnecessary-dunder-call
        exponential_histogram_aggregation_2._value_positive.__init__()
        exponential_histogram_aggregation_2._value_negative.__init__()
        exponential_histogram_aggregation_2._sum = 0
        exponential_histogram_aggregation_2._count = 0
        exponential_histogram_aggregation_2._zero_count = 0
        exponential_histogram_aggregation_2._min = 0
        exponential_histogram_aggregation_2._max = 0
        exponential_histogram_aggregation_2._mapping = LogarithmMapping(
            LogarithmMapping._max_scale
        )

        for attribute in [
            "_value_positive",
            "_value_negative",
            "_sum",
            "_count",
            "_zero_count",
            "_min",
            "_max",
            "_mapping",
        ]:
            setattr(
                exponential_histogram_aggregation_2,
                attribute,
                getattr(exponential_histogram_aggregation_1, attribute),
            )

        self.require_equal(
            exponential_histogram_aggregation_1,
            exponential_histogram_aggregation_2,
        )

    def test_zero_count_by_increment(self):
        now = time_ns()
        ctx = Context()

        exponential_histogram_aggregation_0 = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
            )
        )

        increment = 10

        for _ in range(increment):
            exponential_histogram_aggregation_0.aggregate(
                Measurement(0, now, Mock(), ctx)
            )
        exponential_histogram_aggregation_1 = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
            )
        )

        def mock_increment(self, bucket_index: int) -> None:
            """
            Increments a bucket
            """

            self._counts[bucket_index] += increment

        exponential_histogram_aggregation_1._value_positive = Buckets()

        with patch.object(
            exponential_histogram_aggregation_1._value_positive,
            "increment_bucket",
            MethodType(
                mock_increment,
                exponential_histogram_aggregation_1._value_positive,
            ),
        ):
            exponential_histogram_aggregation_1.aggregate(
                Measurement(0, now, Mock(), ctx)
            )
            exponential_histogram_aggregation_1._count *= increment
            exponential_histogram_aggregation_1._zero_count *= increment

        self.require_equal(
            exponential_histogram_aggregation_0,
            exponential_histogram_aggregation_1,
        )

    def test_one_count_by_increment(self):
        now = time_ns()
        ctx = Context()

        exponential_histogram_aggregation_0 = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
            )
        )

        increment = 10

        for _ in range(increment):
            exponential_histogram_aggregation_0.aggregate(
                Measurement(1, now, Mock(), ctx)
            )
        exponential_histogram_aggregation_1 = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
            )
        )

        def mock_increment(self, bucket_index: int) -> None:
            """
            Increments a bucket
            """

            self._counts[bucket_index] += increment

        exponential_histogram_aggregation_1._value_positive = Buckets()

        with patch.object(
            exponential_histogram_aggregation_1._value_positive,
            "increment_bucket",
            MethodType(
                mock_increment,
                exponential_histogram_aggregation_1._value_positive,
            ),
        ):
            exponential_histogram_aggregation_1.aggregate(
                Measurement(1, now, Mock(), ctx)
            )
            exponential_histogram_aggregation_1._count *= increment
            exponential_histogram_aggregation_1._sum *= increment

        self.require_equal(
            exponential_histogram_aggregation_0,
            exponential_histogram_aggregation_1,
        )

    def test_boundary_statistics(self):
        total = MAX_NORMAL_EXPONENT - MIN_NORMAL_EXPONENT + 1

        for scale in range(
            LogarithmMapping._min_scale, LogarithmMapping._max_scale + 1
        ):
            above = 0
            below = 0

            if scale <= 0:
                mapping = ExponentMapping(scale)
            else:
                mapping = LogarithmMapping(scale)

            for exp in range(MIN_NORMAL_EXPONENT, MAX_NORMAL_EXPONENT + 1):
                value = ldexp(1, exp)

                index = mapping.map_to_index(value)

                with self.assertNotRaises(Exception):
                    boundary = mapping.get_lower_boundary(index + 1)

                if boundary < value:
                    above += 1
                elif boundary > value:
                    below += 1

            self.assertInEpsilon(0.5, above / total, 0.05)
            self.assertInEpsilon(0.5, below / total, 0.06)

    def test_min_max_size(self):
        """
        Tests that the minimum max_size is the right value.
        """

        exponential_histogram_aggregation = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
                max_size=_ExponentialBucketHistogramAggregation._min_max_size,
            )
        )

        # The minimum and maximum normal floating point values are used here to
        # make sure the mapping can contain the full range of values.
        exponential_histogram_aggregation.aggregate(Mock(value=float_info.min))
        exponential_histogram_aggregation.aggregate(Mock(value=float_info.max))

        # This means the smallest max_scale is enough for the full range of the
        # normal floating point values.
        self.assertEqual(
            len(exponential_histogram_aggregation._value_positive._counts),
            exponential_histogram_aggregation._min_max_size,
        )

    def test_aggregate_collect(self):
        """
        Tests a repeated cycle of aggregation and collection.
        """
        now = time_ns()
        ctx = Context()

        exponential_histogram_aggregation = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
            )
        )

        exponential_histogram_aggregation.aggregate(
            Measurement(2, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.collect(
            AggregationTemporality.CUMULATIVE, 0
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(2, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.collect(
            AggregationTemporality.CUMULATIVE, 0
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(2, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.collect(
            AggregationTemporality.CUMULATIVE, 0
        )

    def test_collect_results_cumulative(self) -> None:
        now = time_ns()
        ctx = Context()

        exponential_histogram_aggregation = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
            )
        )
        self.maxDiff = None

        self.assertEqual(exponential_histogram_aggregation._mapping._scale, 20)

        exponential_histogram_aggregation.aggregate(
            Measurement(2, now, Mock(), ctx)
        )
        self.assertEqual(exponential_histogram_aggregation._mapping._scale, 20)

        exponential_histogram_aggregation.aggregate(
            Measurement(4, now, Mock(), ctx)
        )
        self.assertEqual(exponential_histogram_aggregation._mapping._scale, 7)

        exponential_histogram_aggregation.aggregate(
            Measurement(1, now, Mock(), ctx)
        )
        self.assertEqual(exponential_histogram_aggregation._mapping._scale, 6)

        collection_0 = exponential_histogram_aggregation.collect(
            AggregationTemporality.CUMULATIVE, Mock()
        )

        self.assertEqual(len(collection_0.positive.bucket_counts), 160)

        self.assertEqual(collection_0.count, 3)
        self.assertEqual(collection_0.sum, 7)
        self.assertEqual(collection_0.scale, 6)
        self.assertEqual(collection_0.zero_count, 0)
        self.assertEqual(
            collection_0.positive.bucket_counts,
            [1, *[0] * 63, 1, *[0] * 63, 1, *[0] * 31],
        )
        self.assertEqual(collection_0.flags, 0)
        self.assertEqual(collection_0.min, 1)
        self.assertEqual(collection_0.max, 4)

        exponential_histogram_aggregation.aggregate(
            Measurement(1, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(8, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(0.5, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(0.1, now, Mock(), ctx)
        )
        exponential_histogram_aggregation.aggregate(
            Measurement(0.045, now, Mock(), ctx)
        )

        collection_1 = exponential_histogram_aggregation.collect(
            AggregationTemporality.CUMULATIVE, Mock()
        )

        previous_count = collection_1.positive.bucket_counts[0]

        count_counts = [[previous_count, 0]]

        for count in collection_1.positive.bucket_counts:
            if count == previous_count:
                count_counts[-1][1] += 1
            else:
                previous_count = count
                count_counts.append([previous_count, 1])

        self.assertEqual(collection_1.count, 8)
        self.assertEqual(collection_1.sum, 16.645)
        self.assertEqual(collection_1.scale, 4)
        self.assertEqual(collection_1.zero_count, 0)

        self.assertEqual(
            collection_1.positive.bucket_counts,
            [
                1,
                *[0] * 17,
                1,
                *[0] * 36,
                1,
                *[0] * 15,
                2,
                *[0] * 15,
                1,
                *[0] * 15,
                1,
                *[0] * 15,
                1,
                *[0] * 40,
            ],
        )
        self.assertEqual(collection_1.flags, 0)
        self.assertEqual(collection_1.min, 0.045)
        self.assertEqual(collection_1.max, 8)

    def test_cumulative_aggregation_with_random_data(self) -> None:
        now = time_ns()
        ctx = Context()

        histogram = _ExponentialBucketHistogramAggregation(
            Mock(),
            _default_reservoir_factory(_ExponentialBucketHistogramAggregation),
            AggregationTemporality.DELTA,
            Mock(),
        )

        def collect_and_validate(values, histogram) -> None:
            result: ExponentialHistogramDataPoint = histogram.collect(
                AggregationTemporality.CUMULATIVE, 0
            )
            buckets = result.positive.bucket_counts
            scale = result.scale
            index_start = result.positive.offset

            for i in range(len(buckets)):
                index = index_start + i
                count = buckets[i]
                lower_bound = 2 ** (index / (2**scale))
                upper_bound = 2 ** ((index + 1) / (2**scale))
                matches = 0
                for value in values:
                    # pylint: disable=chained-comparison
                    if value > lower_bound and value <= upper_bound:
                        matches += 1
                assert matches == count, (
                    f"index: {index}, count: {count}, scale: {scale}, lower_bound: {lower_bound}, upper_bound: {upper_bound}, matches: {matches}"
                )

            assert sum(buckets) + result.zero_count == len(values)
            assert result.sum == sum(values)
            assert result.count == len(values)
            assert result.min == min(values)
            assert result.max == max(values)
            assert result.zero_count == len([v for v in values if v == 0])
            assert scale >= 3

        seed = randrange(maxsize)
        # This test case is executed with random values every time. In order to
        # run this test case with the same values used in a previous execution,
        # check the value printed by that previous execution of this test case
        # and use the same value for the seed variable in the line below.
        # seed = 3373389994391084876

        random_generator = Random(seed)
        print(f"seed for {currentframe().f_code.co_name} is {seed}")

        values = []
        for i in range(2000):
            # avoid both values being 0
            value = random_generator.randint(0 if i else 1, 1000)
            values.append(value)
            histogram.aggregate(Measurement(value, now, Mock(), ctx))
            if i % 20 == 0:
                collect_and_validate(values, histogram)

        collect_and_validate(values, histogram)

    def test_merge_collect_cumulative(self):
        now = time_ns()
        ctx = Context()

        exponential_histogram_aggregation = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
                max_size=4,
            )
        )

        for value in [2, 4, 8, 16]:
            exponential_histogram_aggregation.aggregate(
                Measurement(value, now, Mock(), ctx)
            )

        self.assertEqual(exponential_histogram_aggregation._mapping.scale, 0)
        self.assertEqual(
            exponential_histogram_aggregation._value_positive.offset, 0
        )
        self.assertEqual(
            exponential_histogram_aggregation._value_positive.counts,
            [1, 1, 1, 1],
        )

        result_0 = exponential_histogram_aggregation.collect(
            AggregationTemporality.CUMULATIVE,
            0,
        )

        self.assertEqual(result_0.scale, 0)

        for value in [1, 2, 4, 8]:
            exponential_histogram_aggregation.aggregate(
                Measurement(1 / value, now, Mock(), ctx)
            )

        self.assertEqual(exponential_histogram_aggregation._mapping.scale, 0)
        self.assertEqual(
            exponential_histogram_aggregation._value_positive.offset, -4
        )
        self.assertEqual(
            exponential_histogram_aggregation._value_positive.counts,
            [1, 1, 1, 1],
        )

        result_1 = exponential_histogram_aggregation.collect(
            AggregationTemporality.CUMULATIVE,
            0,
        )

        self.assertEqual(result_1.scale, -1)

    def test_merge_collect_delta(self):
        now = time_ns()
        ctx = Context()

        exponential_histogram_aggregation = (
            _ExponentialBucketHistogramAggregation(
                Mock(),
                _default_reservoir_factory(
                    _ExponentialBucketHistogramAggregation
                ),
                AggregationTemporality.DELTA,
                Mock(),
                max_size=4,
            )
        )

        for value in [2, 4, 8, 16]:
            exponential_histogram_aggregation.aggregate(
                Measurement(value, now, Mock(), ctx)
            )

        self.assertEqual(exponential_histogram_aggregation._mapping.scale, 0)
        self.assertEqual(
            exponential_histogram_aggregation._value_positive.offset, 0
        )
        self.assertEqual(
            exponential_histogram_aggregation._value_positive.counts,
            [1, 1, 1, 1],
        )

        result = exponential_histogram_aggregation.collect(
            AggregationTemporality.DELTA,
            0,
        )

        for value in [1, 2, 4, 8]:
            exponential_histogram_aggregation.aggregate(
                Measurement(1 / value, now, Mock(), ctx)
            )

        self.assertEqual(exponential_histogram_aggregation._mapping.scale, 0)
        self.assertEqual(
            exponential_histogram_aggregation._value_positive.offset, -4
        )
        self.assertEqual(
            exponential_histogram_aggregation._value_positive.counts,
            [1, 1, 1, 1],
        )

        result_1 = exponential_histogram_aggregation.collect(
            AggregationTemporality.DELTA,
            0,
        )

        self.assertEqual(result.scale, result_1.scale)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from enum import Enum
from typing import Final

from typing_extensions import deprecated

ERROR_MESSAGE: Final = "error.message"
"""
Deprecated: Use domain-specific error message attribute. For example, use `feature_flag.error.message` for feature flag errors.
"""

ERROR_TYPE: Final = "error.type"
"""
Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.error_attributes.ERROR_TYPE`.
"""


@deprecated(
    "Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.error_attributes.ErrorTypeValues`."
)
class ErrorTypeValues(Enum):
    OTHER = "_OTHER"
    """Deprecated in favor of stable :py:const:`opentelemetry.semconv.attributes.error_attributes.ErrorTypeValues.OTHER`."""



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from concurrent.futures import (  # pylint: disable=no-name-in-module
    ThreadPoolExecutor,
)

from opentracing.ext import tags

# pylint: disable=import-error
from ..otel_ot_shim_tracer import MockTracer
from ..testcase import OpenTelemetryTestCase
from ..utils import get_one_by_tag
from .response_listener import ResponseListener


class Client:
    def __init__(self, tracer):
        self.tracer = tracer
        self.executor = ThreadPoolExecutor(max_workers=3)

    def _task(self, message, listener):
        # pylint: disable=no-self-use
        res = f"{message}::response"
        listener.on_response(res)
        return res

    def send_sync(self, message):
        span = self.tracer.start_span("send")
        span.set_tag(tags.SPAN_KIND, tags.SPAN_KIND_RPC_CLIENT)

        listener = ResponseListener(span)
        return self.executor.submit(self._task, message, listener).result()


class TestThreads(OpenTelemetryTestCase):
    def setUp(self):  # pylint: disable=invalid-name
        self.tracer = MockTracer()

    def test_main(self):
        client = Client(self.tracer)
        res = client.send_sync("message")
        self.assertEqual(res, "message::response")

        spans = self.tracer.finished_spans()
        self.assertEqual(len(spans), 1)

        span = get_one_by_tag(spans, tags.SPAN_KIND, tags.SPAN_KIND_RPC_CLIENT)
        self.assertIsNotNone(span)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from opentelemetry import trace
from opentelemetry.context import attach, detach, set_value
from opentelemetry.exporter.opencensus.trace_exporter import (
    OpenCensusSpanExporter,
)
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor
from opentelemetry.test.test_base import TestBase


class ExportStatusSpanProcessor(SimpleSpanProcessor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.export_status = []

    def on_end(self, span):
        token = attach(set_value("suppress_instrumentation", True))
        self.export_status.append(self.span_exporter.export((span,)))
        detach(token)


class TestOpenCensusSpanExporter(TestBase):
    def setUp(self):
        super().setUp()

        trace.set_tracer_provider(TracerProvider())
        self.tracer = trace.get_tracer(__name__)
        self.span_processor = ExportStatusSpanProcessor(
            OpenCensusSpanExporter(endpoint="localhost:55678")
        )

        trace.get_tracer_provider().add_span_processor(self.span_processor)

    def test_export(self):
        with self.tracer.start_as_current_span("foo"):
            with self.tracer.start_as_current_span("bar"):
                with self.tracer.start_as_current_span("baz"):
                    pass

        self.assertTrue(len(self.span_processor.export_status), 3)

        for export_status in self.span_processor.export_status:
            self.assertEqual(export_status.name, "SUCCESS")
            self.assertEqual(export_status.value, 0)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import unittest
from unittest.mock import patch

from opentelemetry import context
from opentelemetry.context.context import Context
from opentelemetry.context.contextvars_context import ContextVarsRuntimeContext
from opentelemetry.environment_variables import OTEL_PYTHON_CONTEXT


def _do_work() -> str:
    key = context.create_key("say")
    context.attach(context.set_value(key, "bar"))
    return key


class TestContext(unittest.TestCase):
    def setUp(self):
        context.attach(Context())

    def test_context_key(self):
        key1 = context.create_key("say")
        key2 = context.create_key("say")
        self.assertNotEqual(key1, key2)
        first = context.set_value(key1, "foo")
        second = context.set_value(key2, "bar")
        self.assertEqual(context.get_value(key1, context=first), "foo")
        self.assertEqual(context.get_value(key2, context=second), "bar")

    def test_context(self):
        key1 = context.create_key("say")
        self.assertIsNone(context.get_value(key1))
        empty = context.get_current()
        second = context.set_value(key1, "foo")
        self.assertEqual(context.get_value(key1, context=second), "foo")

        key2 = _do_work()
        self.assertEqual(context.get_value(key2), "bar")
        third = context.get_current()

        self.assertIsNone(context.get_value(key1, context=empty))
        self.assertEqual(context.get_value(key1, context=second), "foo")
        self.assertEqual(context.get_value(key2, context=third), "bar")

    def test_set_value(self):
        first = context.set_value("a", "yyy")
        second = context.set_value("a", "zzz")
        third = context.set_value("a", "---", first)
        self.assertEqual("yyy", context.get_value("a", context=first))
        self.assertEqual("zzz", context.get_value("a", context=second))
        self.assertEqual("---", context.get_value("a", context=third))
        self.assertEqual(None, context.get_value("a"))

    def test_context_is_immutable(self):
        ctx = context.get_current()
        with self.assertRaises(ValueError):
            ctx["test"] = "cant-change-immutable"
        with self.assertRaises(ValueError):
            del ctx["test"]
        with self.assertRaises(ValueError):
            ctx.pop("test")
        with self.assertRaises(ValueError):
            ctx.popitem()
        with self.assertRaises(ValueError):
            ctx.clear()
        with self.assertRaises(ValueError):
            ctx.update({"test": "cant-change-immutable"})
        with self.assertRaises(ValueError):
            ctx.setdefault("test", "cant-change-immutable")
        with self.assertRaises(ValueError):
            ctx |= {"test": "cant-change-immutable"}

    def test_set_current(self):
        context.attach(context.set_value("a", "yyy"))

        token = context.attach(context.set_value("a", "zzz"))
        self.assertEqual("zzz", context.get_value("a"))

        context.detach(token)
        self.assertEqual("yyy", context.get_value("a"))


class TestInitContext(unittest.TestCase):
    def test_load_runtime_context_default(self):
        ctx = context._load_runtime_context()  # pylint: disable=W0212
        self.assertIsInstance(ctx, ContextVarsRuntimeContext)

    @patch.dict("os.environ", {OTEL_PYTHON_CONTEXT: "contextvars_context"})
    def test_load_runtime_context(self):  # type: ignore[misc]
        ctx = context._load_runtime_context()  # pylint: disable=W0212
        self.assertIsInstance(ctx, ContextVarsRuntimeContext)

    @patch.dict("os.environ", {OTEL_PYTHON_CONTEXT: "foo"})
    def test_load_runtime_context_fallback(self):  # type: ignore[misc]
        ctx = context._load_runtime_context()  # pylint: disable=W0212
        self.assertIsInstance(ctx, ContextVarsRuntimeContext)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Generic, TypeVar

ExporterT = TypeVar("ExporterT")

CUSTOM_HEADERS: dict[str, str] = {
    "x-custom-header": "custom-value",
    "x-another-header": "another-value",
}


@dataclass
class ExporterConfig(Generic[ExporterT]):
    id: str
    exporter_class: type[ExporterT]
    kwargs: dict[str, Any] = field(default_factory=dict)

    def build(self) -> ExporterT:
        return self.exporter_class(**self.kwargs)


def _attrs_to_dict(attributes) -> dict:
    result = {}
    for kv in attributes:
        which = kv.value.WhichOneof("value")
        if which == "string_value":
            result[kv.key] = kv.value.string_value
        elif which == "int_value":
            result[kv.key] = kv.value.int_value
        elif which == "double_value":
            result[kv.key] = kv.value.double_value
        elif which == "bool_value":
            result[kv.key] = kv.value.bool_value
    return result



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0
from django.http import HttpResponse

from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
    ConsoleSpanExporter,
)

trace.set_tracer_provider(TracerProvider())

trace.get_tracer_provider().add_span_processor(
    BatchSpanProcessor(ConsoleSpanExporter())
)


def home_page_view(request):
    return HttpResponse("Hello, world")



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

OTEL_LOGS_EXPORTER = "OTEL_LOGS_EXPORTER"
"""
.. envvar:: OTEL_LOGS_EXPORTER

"""

OTEL_METRICS_EXPORTER = "OTEL_METRICS_EXPORTER"
"""
.. envvar:: OTEL_METRICS_EXPORTER

Specifies which exporter is used for metrics. See `General SDK Configuration
<https://opentelemetry.io/docs/concepts/sdk-configuration/general-sdk-configuration/#otel_metrics_exporter>`_.

**Default value:** ``"otlp"``

**Example:**

``export OTEL_METRICS_EXPORTER="prometheus"``

Accepted values for ``OTEL_METRICS_EXPORTER`` are:

- ``"otlp"``
- ``"prometheus"``
- ``"none"``: No automatically configured exporter for metrics.

.. note::

    Exporter packages may add entry points for group ``opentelemetry_metrics_exporter`` which
    can then be used with this environment variable by name. The entry point should point to
    either a `opentelemetry.sdk.metrics.export.MetricExporter` (push exporter) or
    `opentelemetry.sdk.metrics.export.MetricReader` (pull exporter) subclass; it must be
    constructable without any required arguments. This mechanism is considered experimental and
    may change in subsequent releases.
"""

OTEL_PROPAGATORS = "OTEL_PROPAGATORS"
"""
.. envvar:: OTEL_PROPAGATORS
"""

OTEL_PYTHON_CONTEXT = "OTEL_PYTHON_CONTEXT"
"""
.. envvar:: OTEL_PYTHON_CONTEXT
"""

OTEL_PYTHON_ID_GENERATOR = "OTEL_PYTHON_ID_GENERATOR"
"""
.. envvar:: OTEL_PYTHON_ID_GENERATOR
"""

OTEL_TRACES_EXPORTER = "OTEL_TRACES_EXPORTER"
"""
.. envvar:: OTEL_TRACES_EXPORTER
"""

OTEL_PYTHON_TRACER_PROVIDER = "OTEL_PYTHON_TRACER_PROVIDER"
"""
.. envvar:: OTEL_PYTHON_TRACER_PROVIDER
"""

OTEL_PYTHON_METER_PROVIDER = "OTEL_PYTHON_METER_PROVIDER"
"""
.. envvar:: OTEL_PYTHON_METER_PROVIDER
"""

_OTEL_PYTHON_LOGGER_PROVIDER = "OTEL_PYTHON_LOGGER_PROVIDER"
"""
.. envvar:: OTEL_PYTHON_LOGGER_PROVIDER
"""



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from enum import Enum
from typing import Final

from typing_extensions import deprecated

GEN_AI_AGENT_DESCRIPTION: Final = "gen_ai.agent.description"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_AGENT_ID: Final = "gen_ai.agent.id"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_AGENT_NAME: Final = "gen_ai.agent.name"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_AGENT_VERSION: Final = "gen_ai.agent.version"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_COMPLETION: Final = "gen_ai.completion"
"""
Deprecated: Removed, no replacement at this time.
"""

GEN_AI_CONVERSATION_ID: Final = "gen_ai.conversation.id"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_DATA_SOURCE_ID: Final = "gen_ai.data_source.id"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_EMBEDDINGS_DIMENSION_COUNT: Final = "gen_ai.embeddings.dimension.count"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_EVALUATION_EXPLANATION: Final = "gen_ai.evaluation.explanation"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_EVALUATION_NAME: Final = "gen_ai.evaluation.name"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_EVALUATION_SCORE_LABEL: Final = "gen_ai.evaluation.score.label"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_EVALUATION_SCORE_VALUE: Final = "gen_ai.evaluation.score.value"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_INPUT_MESSAGES: Final = "gen_ai.input.messages"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_OPENAI_REQUEST_RESPONSE_FORMAT: Final = (
    "gen_ai.openai.request.response_format"
)
"""
Deprecated: Replaced by `gen_ai.output.type`, which has moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_OPENAI_REQUEST_SEED: Final = "gen_ai.openai.request.seed"
"""
Deprecated: Replaced by `gen_ai.request.seed`, which has moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_OPENAI_REQUEST_SERVICE_TIER: Final = (
    "gen_ai.openai.request.service_tier"
)
"""
Deprecated: Replaced by `openai.request.service_tier`, which has moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_OPENAI_RESPONSE_SERVICE_TIER: Final = (
    "gen_ai.openai.response.service_tier"
)
"""
Deprecated: Replaced by `openai.response.service_tier`, which has moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_OPENAI_RESPONSE_SYSTEM_FINGERPRINT: Final = (
    "gen_ai.openai.response.system_fingerprint"
)
"""
Deprecated: Replaced by `openai.response.system_fingerprint`, which has moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_OPERATION_NAME: Final = "gen_ai.operation.name"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_OUTPUT_MESSAGES: Final = "gen_ai.output.messages"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_OUTPUT_TYPE: Final = "gen_ai.output.type"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_PROMPT: Final = "gen_ai.prompt"
"""
Deprecated: Removed, no replacement at this time.
"""

GEN_AI_PROMPT_NAME: Final = "gen_ai.prompt.name"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_PROVIDER_NAME: Final = "gen_ai.provider.name"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_REQUEST_CHOICE_COUNT: Final = "gen_ai.request.choice.count"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_REQUEST_ENCODING_FORMATS: Final = "gen_ai.request.encoding_formats"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_REQUEST_FREQUENCY_PENALTY: Final = "gen_ai.request.frequency_penalty"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_REQUEST_MAX_TOKENS: Final = "gen_ai.request.max_tokens"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_REQUEST_MODEL: Final = "gen_ai.request.model"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_REQUEST_PRESENCE_PENALTY: Final = "gen_ai.request.presence_penalty"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_REQUEST_SEED: Final = "gen_ai.request.seed"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_REQUEST_STOP_SEQUENCES: Final = "gen_ai.request.stop_sequences"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_REQUEST_STREAM: Final = "gen_ai.request.stream"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_REQUEST_TEMPERATURE: Final = "gen_ai.request.temperature"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_REQUEST_TOP_K: Final = "gen_ai.request.top_k"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_REQUEST_TOP_P: Final = "gen_ai.request.top_p"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_RESPONSE_FINISH_REASONS: Final = "gen_ai.response.finish_reasons"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_RESPONSE_ID: Final = "gen_ai.response.id"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_RESPONSE_MODEL: Final = "gen_ai.response.model"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_RESPONSE_TIME_TO_FIRST_CHUNK: Final = (
    "gen_ai.response.time_to_first_chunk"
)
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_RETRIEVAL_DOCUMENTS: Final = "gen_ai.retrieval.documents"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_RETRIEVAL_QUERY_TEXT: Final = "gen_ai.retrieval.query.text"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_SYSTEM: Final = "gen_ai.system"
"""
Deprecated: Replaced by `gen_ai.provider.name`, which has moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_SYSTEM_INSTRUCTIONS: Final = "gen_ai.system_instructions"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_TOKEN_TYPE: Final = "gen_ai.token.type"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_TOOL_CALL_ARGUMENTS: Final = "gen_ai.tool.call.arguments"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_TOOL_CALL_ID: Final = "gen_ai.tool.call.id"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_TOOL_CALL_RESULT: Final = "gen_ai.tool.call.result"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_TOOL_DEFINITIONS: Final = "gen_ai.tool.definitions"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_TOOL_DESCRIPTION: Final = "gen_ai.tool.description"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_TOOL_NAME: Final = "gen_ai.tool.name"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_TOOL_TYPE: Final = "gen_ai.tool.type"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_USAGE_CACHE_CREATION_INPUT_TOKENS: Final = (
    "gen_ai.usage.cache_creation.input_tokens"
)
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_USAGE_CACHE_READ_INPUT_TOKENS: Final = (
    "gen_ai.usage.cache_read.input_tokens"
)
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_USAGE_COMPLETION_TOKENS: Final = "gen_ai.usage.completion_tokens"
"""
Deprecated: Replaced by `gen_ai.usage.output_tokens`, which has moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_USAGE_INPUT_TOKENS: Final = "gen_ai.usage.input_tokens"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_USAGE_OUTPUT_TOKENS: Final = "gen_ai.usage.output_tokens"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_USAGE_PROMPT_TOKENS: Final = "gen_ai.usage.prompt_tokens"
"""
Deprecated: Replaced by `gen_ai.usage.input_tokens`, which has moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_USAGE_REASONING_OUTPUT_TOKENS: Final = (
    "gen_ai.usage.reasoning.output_tokens"
)
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""

GEN_AI_WORKFLOW_NAME: Final = "gen_ai.workflow.name"
"""
Deprecated: Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai).
"""


@deprecated(
    "The attribute gen_ai.openai.request.response_format is deprecated - Replaced by `gen_ai.output.type`, which has moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai)"
)
class GenAiOpenaiRequestResponseFormatValues(Enum):
    TEXT = "text"
    """Text response format."""
    JSON_OBJECT = "json_object"
    """JSON object response format."""
    JSON_SCHEMA = "json_schema"
    """JSON schema response format."""


@deprecated(
    "The attribute gen_ai.openai.request.service_tier is deprecated - Replaced by `openai.request.service_tier`, which has moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai)"
)
class GenAiOpenaiRequestServiceTierValues(Enum):
    AUTO = "auto"
    """The system will utilize scale tier credits until they are exhausted."""
    DEFAULT = "default"
    """The system will utilize the default scale tier."""


@deprecated(
    "The attribute gen_ai.operation.name is deprecated - Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai)"
)
class GenAiOperationNameValues(Enum):
    CHAT = "chat"
    """Chat completion operation such as [OpenAI Chat API](https://platform.openai.com/docs/api-reference/chat)."""
    GENERATE_CONTENT = "generate_content"
    """Multimodal content generation operation such as [Gemini Generate Content](https://ai.google.dev/api/generate-content)."""
    TEXT_COMPLETION = "text_completion"
    """Text completions operation such as [OpenAI Completions API (Legacy)](https://platform.openai.com/docs/api-reference/completions)."""
    EMBEDDINGS = "embeddings"
    """Embeddings operation such as [OpenAI Create embeddings API](https://platform.openai.com/docs/api-reference/embeddings/create)."""
    RETRIEVAL = "retrieval"
    """Retrieval operation such as [OpenAI Search Vector Store API](https://platform.openai.com/docs/api-reference/vector-stores/search)."""
    CREATE_AGENT = "create_agent"
    """Create GenAI agent."""
    INVOKE_AGENT = "invoke_agent"
    """Invoke GenAI agent."""
    EXECUTE_TOOL = "execute_tool"
    """Execute a tool."""
    INVOKE_WORKFLOW = "invoke_workflow"
    """Invoke GenAI workflow."""


@deprecated(
    "The attribute gen_ai.output.type is deprecated - Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai)"
)
class GenAiOutputTypeValues(Enum):
    TEXT = "text"
    """Plain text."""
    JSON = "json"
    """JSON object with known or unknown schema."""
    IMAGE = "image"
    """Image."""
    SPEECH = "speech"
    """Speech."""


@deprecated(
    "The attribute gen_ai.provider.name is deprecated - Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai)"
)
class GenAiProviderNameValues(Enum):
    OPENAI = "openai"
    """[OpenAI](https://openai.com/)."""
    GCP_GEN_AI = "gcp.gen_ai"
    """Any Google generative AI endpoint."""
    GCP_VERTEX_AI = "gcp.vertex_ai"
    """[Vertex AI](https://cloud.google.com/vertex-ai)."""
    GCP_GEMINI = "gcp.gemini"
    """[Gemini](https://cloud.google.com/products/gemini)."""
    ANTHROPIC = "anthropic"
    """[Anthropic](https://www.anthropic.com/)."""
    COHERE = "cohere"
    """[Cohere](https://cohere.com/)."""
    AZURE_AI_INFERENCE = "azure.ai.inference"
    """Azure AI Inference."""
    AZURE_AI_OPENAI = "azure.ai.openai"
    """[Azure OpenAI](https://learn.microsoft.com/en-us/azure/ai-services/openai/overview)."""
    IBM_WATSONX_AI = "ibm.watsonx.ai"
    """[IBM Watsonx AI](https://www.ibm.com/products/watsonx-ai)."""
    AWS_BEDROCK = "aws.bedrock"
    """[AWS Bedrock](https://aws.amazon.com/bedrock)."""
    PERPLEXITY = "perplexity"
    """[Perplexity](https://www.perplexity.ai/)."""
    X_AI = "x_ai"
    """[xAI](https://x.ai/)."""
    DEEPSEEK = "deepseek"
    """[DeepSeek](https://www.deepseek.com/)."""
    GROQ = "groq"
    """[Groq](https://groq.com/)."""
    MISTRAL_AI = "mistral_ai"
    """[Mistral AI](https://mistral.ai/)."""


@deprecated(
    "The attribute gen_ai.system is deprecated - Replaced by `gen_ai.provider.name`, which has moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai)"
)
class GenAiSystemValues(Enum):
    OPENAI = "openai"
    """OpenAI."""
    GCP_GEN_AI = "gcp.gen_ai"
    """Any Google generative AI endpoint."""
    GCP_VERTEX_AI = "gcp.vertex_ai"
    """Vertex AI."""
    GCP_GEMINI = "gcp.gemini"
    """Gemini."""
    VERTEX_AI = "vertex_ai"
    """Deprecated: Replaced by `gcp.vertex_ai`."""
    GEMINI = "gemini"
    """Deprecated: Replaced by `gcp.gemini`."""
    ANTHROPIC = "anthropic"
    """Anthropic."""
    COHERE = "cohere"
    """Cohere."""
    AZ_AI_INFERENCE = "az.ai.inference"
    """Deprecated: Replaced by `azure.ai.inference`."""
    AZ_AI_OPENAI = "az.ai.openai"
    """Deprecated: Replaced by `azure.ai.openai`."""
    AZURE_AI_INFERENCE = "azure.ai.inference"
    """Azure AI Inference."""
    AZURE_AI_OPENAI = "azure.ai.openai"
    """Azure OpenAI."""
    IBM_WATSONX_AI = "ibm.watsonx.ai"
    """IBM Watsonx AI."""
    AWS_BEDROCK = "aws.bedrock"
    """AWS Bedrock."""
    PERPLEXITY = "perplexity"
    """Perplexity."""
    XAI = "xai"
    """xAI."""
    DEEPSEEK = "deepseek"
    """DeepSeek."""
    GROQ = "groq"
    """Groq."""
    MISTRAL_AI = "mistral_ai"
    """Mistral AI."""


@deprecated(
    "The attribute gen_ai.token.type is deprecated - Moved to the [OpenTelemetry GenAI semantic conventions repository](https://github.com/open-telemetry/semantic-conventions-genai)"
)
class GenAiTokenTypeValues(Enum):
    INPUT = "input"
    """Input tokens (prompt, input, etc.)."""
    COMPLETION = "output"
    """Deprecated: Replaced by `output`."""
    OUTPUT = "output"
    """Output tokens (completion, response, etc.)."""



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

"""
This library allows exporting OpenTelemetry telemetry to a file or stream
in OTLP JSON (JSON Lines) format.

Usage
-----

The **OTLP JSON File Exporter** writes `OpenTelemetry`_ traces, metrics, and
logs as compact JSON Lines to a file, an arbitrary ``IO[str]`` stream or
``sys.stdout``.

Three exporters are provided:

- :class:`~opentelemetry.exporter.otlp.json.file.trace_exporter.FileSpanExporter` - traces
- :class:`~opentelemetry.exporter.otlp.json.file.metric_exporter.FileMetricExporter` - metrics
- :class:`~opentelemetry.exporter.otlp.json.file._log_exporter.FileLogExporter` - logs

Each exporter accepts a destination in one of three ways:

- ``path`` - a file path opened in append mode (UTF-8)
- ``stream`` - any ``IO[str]`` text stream
- neither - defaults to ``sys.stdout``

.. _OpenTelemetry: https://github.com/open-telemetry/opentelemetry-python/

.. code:: python

    from opentelemetry import trace
    from opentelemetry.exporter.otlp.json.file import FileSpanExporter
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.sdk.trace.export import BatchSpanProcessor

    # Write traces to a file (append mode, UTF-8)
    exporter = FileSpanExporter("/var/log/otel/traces.jsonl")

    # Alternatively, write to an explicit stream:
    # exporter = FileSpanExporter(stream=open("traces.jsonl", "a"))

    # Or write to sys.stdout (default when no path or stream is given):
    # exporter = FileSpanExporter()

    trace.set_tracer_provider(TracerProvider())
    trace.get_tracer_provider().add_span_processor(
        BatchSpanProcessor(exporter)
    )

    tracer = trace.get_tracer(__name__)
    with tracer.start_as_current_span("foo"):
        print("Hello world!")

The metric exporter supports the following environment variables:

- :envvar:`OTEL_EXPORTER_OTLP_METRICS_TEMPORALITY_PREFERENCE`
- :envvar:`OTEL_EXPORTER_OTLP_METRICS_DEFAULT_HISTOGRAM_AGGREGATION`

API
---
"""

from opentelemetry.exporter.otlp.json.file.metric_exporter import (
    FileMetricExporter,
)
from opentelemetry.exporter.otlp.json.file.trace_exporter import (
    FileSpanExporter,
)

__all__ = [
    "FileMetricExporter",
    "FileSpanExporter",
]



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import unittest

import opentelemetry.trace as trace_api


# pylint: disable=C0103
class OpenTelemetryTestCase(unittest.TestCase):
    def assertSameTrace(self, spanA, spanB):
        return self.assertEqual(spanA.context.trace_id, spanB.context.trace_id)

    def assertNotSameTrace(self, spanA, spanB):
        return self.assertNotEqual(
            spanA.context.trace_id, spanB.context.trace_id
        )

    def assertIsChildOf(self, spanA, spanB):
        # spanA is child of spanB
        self.assertIsNotNone(spanA.parent)

        ctxA = spanA.parent
        if not isinstance(ctxA, trace_api.SpanContext):
            ctxA = spanA.parent.context

        ctxB = spanB
        if not isinstance(ctxB, trace_api.SpanContext):
            ctxB = spanB.context

        return self.assertEqual(ctxA.span_id, ctxB.span_id)

    def assertIsNotChildOf(self, spanA, spanB):
        # spanA is NOT child of spanB
        if spanA.parent is None:
            return

        ctxA = spanA.parent
        if not isinstance(ctxA, trace_api.SpanContext):
            ctxA = spanA.parent.context

        ctxB = spanB
        if not isinstance(ctxB, trace_api.SpanContext):
            ctxB = spanB.context

        self.assertNotEqual(ctxA.span_id, ctxB.span_id)

    def assertNamesEqual(self, spans, names):
        self.assertEqual(list(map(lambda x: x.name, spans)), names)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

"""Top-level orchestrator for declarative SDK configuration.

Takes a parsed ``OpenTelemetryConfiguration`` and applies it by calling
each per-signal ``configure_*`` factory in order. This is the single
entry point for "apply this config" on the declarative path.
"""

from __future__ import annotations

from logging import getLogger

from opentelemetry.configuration._logger_provider import (
    configure_logger_provider,
)
from opentelemetry.configuration._meter_provider import (
    configure_meter_provider,
)
from opentelemetry.configuration._propagator import configure_propagator
from opentelemetry.configuration._resource import create_resource
from opentelemetry.configuration._tracer_provider import (
    configure_tracer_provider,
)
from opentelemetry.configuration.instrumentation import (
    configure_instrumentation,
)
from opentelemetry.configuration.models import OpenTelemetryConfiguration

_logger = getLogger(__name__)


def configure_sdk(config: OpenTelemetryConfiguration) -> None:
    """Configure the global SDK from a parsed declarative configuration.

    Builds a :class:`Resource` from ``config.resource`` and applies it to
    each signal provider. Sets the global tracer provider, meter provider,
    logger provider, and text map propagator from their respective config
    sections. Sections absent from the config (``None``) leave the
    corresponding global untouched — matching the spec's "noop default"
    behavior.

    Honors the top-level ``disabled`` flag: when true, no globals are set.

    Args:
        config: Parsed ``OpenTelemetryConfiguration`` (typically from
            ``load_config_file``).

    Example:
        >>> from opentelemetry.configuration.file import (
        ...     load_config_file, configure_sdk,
        ... )
        >>> config = load_config_file("otel-config.yaml")
        >>> configure_sdk(config)
    """
    if config.disabled:
        _logger.warning(
            "Declarative configuration has disabled=true; skipping SDK setup."
        )
        return

    resource = create_resource(config.resource)
    configure_tracer_provider(config.tracer_provider, resource)
    configure_meter_provider(config.meter_provider, resource)
    configure_logger_provider(config.logger_provider, resource)
    configure_propagator(config.propagator)
    configure_instrumentation(config.instrumentation_development)
